import globalPluginHandler
import addonHandler
import gui
import os
import wx
from logHandler import log
import datetime
import ui
import config
import api
import vision
from locationHelper import RectLTRB
from logHandler import log
from .demoHandler import demoHandler
import globalVars
try:
	import vision
except ImportError:
	raise RuntimeError("No vision framework support")
from . import isUSARunning, libUSA

def skipEventAndCall(handler):
	def wrapWithEventSkip(event):
		if event:
			event.Skip()
		return handler()
	return wrapWithEventSkip

#We need to initialize translation and localization support:
addonHandler.initTranslation()

class GlobalPlugin(globalPluginHandler.GlobalPlugin):
	"""
	USA globalPlugin class of which an instance is created, i.e. upon loading of NVDA.
	This is used to initialize USA Remote if applicable. Also it sets the help file direction in the settings.
	For the Object Reader we also need the mouseMove event.
	"""

	def isUSAEnabled(self):
		veps = vision.handler.getActiveProviderInstances()
		for vep in veps:
			try:
				if vep.getSettings().getId().lower().strip() == "usa":
					return True
			except:
				pass
		return False

	def enableUSA(self):
		if self.isUSAEnabled(): return
		vepsInfo = vision.handler.getProviderList()
		usaProviderInfo =None
		for vep in vepsInfo:
			if vep.providerId == 'usa':
				usaProviderInfo = vep

		if usaProviderInfo:
			vision.handler.initializeProvider(usaProviderInfo)
		else:
			log.error("USA provider not found! Could not enable provider")

	def disableUSA(self):
		if not self.isUSAEnabled(): return
		vepsInfo = vision.handler.getProviderList()
		usaProviderInfo = None
		for vep in vepsInfo:
			if vep.providerId == 'usa':
				usaProviderInfo = vep

		if usaProviderInfo:
			vision.handler.terminateProvider(usaProviderInfo)
		else:
			log.error("USA provider not found! could not disable provider")

	def getLicenseDialog(self):
		from .licenseDialog import licenseDialog 
		return licenseDialog.USALicenseDialog
		# get the license dialog so that the vision enhancement provider can use it.

	def licenseUSA(self):
		gui.mainFrame.prePopup()
		licenseDialog = self.getLicenseDialog()
		with licenseDialog(gui.mainFrame) as dlg:
			dlg.ShowModal()
		gui.mainFrame.postPopup()

	def USASimpleSettings(self):
		from . import simpleSettings
		gui.mainFrame.prePopup()
		with simpleSettings.USASimpleSettings(gui.mainFrame) as dlg:
			dlg.ShowModal()
		gui.mainFrame.postPopup()

	def USAAdvancedSettings(self):
		gui.mainFrame.prePopup()
		def onVisionCommand():
			gui.mainFrame.popupSettingsDialog(gui.NVDASettingsDialog, gui.settingsDialogs.VisionSettingsPanel)

		wx.CallAfter(onVisionCommand)
		gui.mainFrame.postPopup()

	def updateMenu(self, evt):
		# self.enableUSA_menuitem.Enable(not self.isUSAEnabled())
		# self.disableUSA_menuitem.Enable(self.isUSAEnabled())
		self.quickUSA_menuitem.Enable(self.isUSAEnabled())
		self.advanceUSA_menuitem.Enable(self.isUSAEnabled())

	def create_menu(self):
		self.menu = wx.Menu()

		self.quickUSA_menuitem = self.menu.Append(wx.ID_ANY, _("quick settings USA"))

		gui.mainFrame.sysTrayIcon.Bind(
			wx.EVT_MENU,
			skipEventAndCall(self.USASimpleSettings), self.quickUSA_menuitem
		)

		self.licenseUSA_menuItem = self.menu.Append(wx.ID_ANY, _("License USA"))

		gui.mainFrame.sysTrayIcon.Bind(
			wx.EVT_MENU,
			skipEventAndCall(self.licenseUSA), self.licenseUSA_menuItem
		)
		self.help_item = self.menu.Append(wx.ID_ANY, _("&Ultimate Screen Access help"),
																   _("Manual for Ultimate Screen Access"))
		gui.mainFrame.sysTrayIcon.Bind(wx.EVT_MENU,
									   lambda e: os.startfile(addonHandler.getCodeAddon().getDocFilePath()),
									   self.help_item)

		self.advanceUSA_menuitem = self.menu.Append(wx.ID_ANY, _("advanced settings USA"))

		gui.mainFrame.sysTrayIcon.Bind(
			wx.EVT_MENU,
			skipEventAndCall(self.USAAdvancedSettings), self.advanceUSA_menuitem
		)

		self.submenu_item = gui.mainFrame.sysTrayIcon.menu.Insert(2, wx.ID_ANY, _("Ultimate Screen Access"), self.menu)
		gui.mainFrame.sysTrayIcon.Bind(wx.EVT_MENU_HIGHLIGHT_ALL, self.updateMenu)

	def automaticActivation(self):
		return
		# kicks off the automatic activation protocol at the start of the setup, first it checks if an oldInstance is still
		# activated and tries to deactivate that. After that it activates the current instance
		try:

			license = config.conf['usaAutoActivationLicenseKey']
			product_id = int(config.conf['usaAutoActivationProductID'])

		except:
			wx.CallAfter(gui.messageBox, "could not preform automatic activation! No valid data found on startup!")
			return

		oldInstanceID = config.conf.get("usaAutoActivationOldID", "")
		if len(oldInstanceID) > 0:
			from . import deactivateLicenseDirectly
			try:
				deactivateLicenseDirectly.deactivate(oldInstanceID, license, product_id)
			except:
				log.error("could not deactivate old license!")


		from .licenseDialog import licenseDialog
		from .licenseExtension import licenseExtension
		licenseDialog.ActivateLicense(license)
		config.conf["usaAutoActivationOldID"] = licenseExtension.getInstanceID(product_id)

	def automaticDeactivation(self):
		return
		# if USA closes nicely, we can kick off the deactivation script.
		from .licenseDialog import licenseDialog
		licenseDialog.DeactivateLicense()
		config.conf["usaAutoActivationOldID"] = ""

	def checkDemoAndRunning(self):
		if globalVars.appArgs.secure:
			# should not show starting dialogs if in a secure desktop enviroment.
			return
		if self.checkForOtherPopups():
			wx.CallLater(1000, self.checkDemoAndRunning)
			return
		licenseDialog = self.getLicenseDialog()
		with licenseDialog(gui.mainFrame) as dlg:
			if dlg.shouldShowAtStart:
				gui.mainFrame.prePopup()
				dlg.ShowModal()
				gui.mainFrame.postPopup()
			else:
				log.debugWarning("not showing license dialog on start")

		wx.CallLater(1000, self.checkRunning)
		
	def checkRunning(self):
		if self.checkForOtherPopups():
			wx.CallLater(1000, self.checkRunning)
			return
		with isUSARunning.isUSARunning(gui.mainFrame) as dlg:
			
			if dlg.shouldAsk():			
				gui.mainFrame.prePopup()
				dlg.ShowModal()
				gui.mainFrame.postPopup()
			else:
				log.debugWarning("not showing start USA dialog on start")
    
	def checkForOtherPopups(self):
		if len(gui.mainFrame.Children) > 0:
			return True
		else:
			return False
		

	def __init__(self):
		super().__init__()
		if config.conf.get('usaAutoActivation') == "False": config.conf['usaAutoActivation'] = False
		if config.conf.get('usaAutoActivation', False):
			self.automaticActivation()

		self.create_menu()
		self.remoteHandler = None
		wx.CallAfter(self.initializeRemote)
		# call isLicensed function to make sure that
		wx.CallAfter(libUSA.libUSA.IsLicensed)
		wx.CallLater(1000, self.checkDemoAndRunning)
		log.info("USA: plugin initialized")
  
	def terminate(self):
		if config.conf.get('usaAutoActivation', False):
			self.automaticDeactivation()

		if self.remoteHandler:
			self.remoteHandler.terminate()
			self.remoteHandler = None
		gui.mainFrame.sysTrayIcon.helpMenu.Remove(self.help_item)
		self.help_item.Destroy()
		self.help_item=None

		super().terminate()

	def initializeRemote(self):
		try:
			self.remoteHandler = RemoteHandler()
		except (NotImplementedError, RuntimeError):
			log.info("USA: USA remote capabilities not available")
			self.remoteHandler = None
		else:
			log.info("USA: USA remote capabilities enabled")

class RemoteHandler(object):

	def __init__(self):
		self.connectionType = None
		self.transport = None
		# check if unicorn is running
		try:
			self.remoteClient = next(p for p in globalPluginHandler.runningPlugins if p.__module__=='globalPlugins.unicorn')
		except StopIteration:
			raise RuntimeError
		# register callbacks
		self.remoteClient.callback_manager.register_callback('transport_connected',self.onTransportConnected)
		self.remoteClient.callback_manager.register_callback('transport_disconnected',self.onTransportDisconnected)

		if self.remoteClient.slave_transport:
			self.onTransportConnected(connection_type='slave',transport=self.remoteClient.slave_transport)
		elif self.remoteClient.master_transport:
			self.onTransportConnected(connection_type='master',transport=self.remoteClient.master_transport)

	def sendVisionCommandOverUnicorn(self, command, **kwargs):
		if self.transport:
			self.transport.send(type="vision", command=command, **kwargs)
		else:
			log.error("Unicorn transport for USA not initialized.")

	def terminate(self):
		self.unregisterRemoteTransport()
		self.remoteClient.callback_manager.unregister_callback('transport_connected',self.onTransportConnected)
		self.remoteClient.callback_manager.unregister_callback('transport_disconnected',self.onTransportDisconnected)
		self.remoteClient = None

	def unregisterRemoteTransport(self):
		if self.transport:
			try:
				self.transport.callback_manager.unregister_callback('msg_vision', self.handleVisionPacket)
			except ValueError:
				pass


	def onTransportConnected(self,connection_type,transport,*args,**kwargs):
		log.debugWarning("unicorn/USA: onTransportConnected")
		transport.callback_manager.register_callback('msg_vision', self.handleVisionPacket)
		self.connectionType = connection_type
		self.transport = transport

	def onTransportDisconnected(self,*args,**kwargs):
		# We do not have to unregister callbacks here
		log.debugWarning("unicorn/USA: onTransportDisconnected")
		if self.transport: self.transport.callback_manager.unregister_callback('msg_vision', self.handleVisionPacket)
		self.connectionType = None
		self.transport = None

	def handleVisionPacket(self,command, *args,**kwargs):
		if not vision.handler:
			return
		try:
			usa = vision.handler.getProviderInstance(vision.handler.getProviderInfo("usa"))
		except LookupError:
			usa = None
		if not usa:
			return
		commandMapping = usa.getUnicornCommandMapping()
		kwargs = usa.deserializeUnicornArguments(command, **kwargs)
		log.info(f"{command} {kwargs}")
		try:
			if kwargs:
				commandMapping[command](**kwargs)
			else:
				commandMapping[command]()
		except Exception as e:
			log.error(f"An error occured when handling a vision package for USA: \n {e}")
