"""
Module with a whole bunch of matrix operations used to simplify the creation of the custom color scheme
"""


e  = 2.71828182845
pi = 3.14159265359

inverseMatrix = [
    [-1, 0, 0, 0, 0],
    [0, -1, 0, 0, 0],
    [0, 0, -1, 0, 0],
    [0, 0, 0, 1, 0],
    [1, 1, 1, 0, 1]
]

def zeros_matrix(rows, cols):
    """
    Creates a matrix filled with zeros.
        :param rows: the number of rows the matrix should have
        :param cols: the number of columns the matrix should have

        :return: list of lists that form the matrix
    """
    M = []
    while len(M) < rows:
        M.append([])
        while len(M[-1]) < cols:
            M[-1].append(0.0)

    return M

def matrix_multiply(A, B):
    """
    Returns the product of the matrix A * B
        :param A: The first matrix - ORDER MATTERS!
        :param B: The second matrix

        :return: The product of the two matrices
    """
    # Section 1: Ensure A & B dimensions are correct for multiplication
    rowsA = len(A)
    colsA = len(A[0])
    rowsB = len(B)
    colsB = len(B[0])
    if colsA != rowsB:
        raise ArithmeticError(
            'Number of A columns must equal number of B rows.')

    # Section 2: Store matrix multiplication in a new matrix
    C = zeros_matrix(rowsA, colsB)
    for i in range(rowsA):
        for j in range(colsB):
            total = 0
            for ii in range(colsA):
                total += A[i][ii] * B[ii][j]
            C[i][j] = total

    return C

def matrix_add(A, B):
    rowsA = len(A)
    colsA = len(A[0])
    rowsB = len(B)
    colsB = len(B[0])
    if colsA != colsA or rowsA != rowsB:
        raise ArithmeticError(
            'A not the same size as B.')

    C = zeros_matrix(rowsA, colsA)
    for ii in range(rowsA):
        for jj in range(colsA):
            C[ii][jj] = A[ii][jj] + B[ii][jj]
    return C

def matrix_multiply_scalar(A, c):
    rowsA = len(A)
    colsA = len(A[0])
    C = zeros_matrix(rowsA, colsA)
    for ii in range(rowsA):
        for jj in range(colsA):
            C[ii][jj] = A[ii][jj]*c
    return C

def cos(theta):
    # I could import the math module, but I took Complex Analysis class and that should not go to waste :)
    return (e**(theta*1j)).real
def sin(theta):
    return (e**(theta*1j)).imag

def rotateMatrixX(theta):
    matX = [[1, 0, 0],
            [0, cos(theta), -sin(theta)],
            [0, sin(theta), cos(theta)]]
    return matX

def rotateMatrixY(theta):
    matY = [[cos(theta), 0, -sin(theta)],
            [0, 1, 0],
            [sin(theta), 0, cos(theta)]]
    return matY

def rotateMatrixZ(theta):
    matZ = [[cos(theta), -sin(theta), 0],
            [sin(theta), cos(theta), 0],
            [0, 0, 1]]
    return matZ

def sqrt(a):
    return a**0.5

# alternative: https://docs.microsoft.com/en-us/windows/win32/direct2d/hue-rotate
# https://stackoverflow.com/questions/8507885/shift-hue-of-an-rgb-color/8510751#8510751
def rotateHue(theta):
    cosA = cos(theta)
    sinA = sin(theta)

    matrix = zeros_matrix(3, 3)

    matrix[0][0] = cosA + (1.0 - cosA) / 3.0
    matrix[0][1] = 1. / 3. * (1.0 - cosA) - sqrt(1. / 3.) * sinA
    matrix[0][2] = 1. / 3. * (1.0 - cosA) + sqrt(1. / 3.) * sinA
    matrix[1][0] = 1. / 3. * (1.0 - cosA) + sqrt(1. / 3.) * sinA
    matrix[1][1] = cosA + 1. / 3. * (1.0 - cosA)
    matrix[1][2] = 1. / 3. * (1.0 - cosA) - sqrt(1. / 3.) * sinA
    matrix[2][0] = 1. / 3. * (1.0 - cosA) - sqrt(1. / 3.) * sinA
    matrix[2][1] = 1. / 3. * (1.0 - cosA) + sqrt(1. / 3.) * sinA
    matrix[2][2] = cosA + 1. / 3. * (1.0 - cosA)

    return matrix