# -*- coding: utf-8 -*-
"""
Babbage B.V., 2019-2023

visionEnhancmentProviders/_base/dialogs.py
Contains classes with dialogs and settings. Also the auto-filled vision-settings panel in the general NVDA settings is
somewhat managed here.
"""


import wx # type: ignore

import globalPluginHandler
from gui import guiHelper, nvdaControls
import gui
import addonHandler
import ui
from logHandler import log

from .. import MatrixOperations as mo
addonHandler.initTranslation()


class USASettingsPanel(gui.settingsDialogs.VisionProviderSubPanel_Wrapper):


	def _createProviderSettings(self):
		res = super()._createProviderSettings()
		if not res:
			return res
		settings = self._providerSettings.getSettings()
		conf = self._providerSettings.getSettings().conf
		sHelper = guiHelper.BoxSizerHelper(self, sizer=self._providerSettingsSizer)


		self.thicknessCtrl = sHelper.addLabeledControl(
			_('Wordfollower rectangle thickness'),
			nvdaControls.SelectOnFocusSpinCtrl,
			min=1,
			max=10,
			initial=int(settings.conf.get('caretThickness')) if settings.conf.get('caretThickness') else 2,
		)

		self.objectReaderMode = sHelper.addItem(wx.CheckBox(self, label=_("Activate Object Reader Mode")))
		self.objectReaderMode.SetValue(settings.objectReaderMode )

		self.isRemote = sHelper.addItem(wx.CheckBox(self, label=_("Run USA in remote Mode (with unicorn)")))
		self.isRemote.SetValue(settings.isRemote)

		licenseButton = sHelper.addItem(wx.Button(self, label=_("Manage USA l&icense...")))
		licenseButton.Bind(wx.EVT_BUTTON,self.onLicense)



		# Color scheme selection
		label = _("Color scheme settings")

		BoxSizerHelper = wx.StaticBoxSizer(wx.VERTICAL, self, label=label)
		colorschemeBox = BoxSizerHelper.GetStaticBox()
		colorschemeGroup = guiHelper.BoxSizerHelper(self, sizer=BoxSizerHelper)
		sHelper.addItem(colorschemeGroup)


		modifierBoxLabel = _("&Select Colorschemes")
		colorSchemeDict = settings._get_allColorschemes()

		self.modifierChoices = [ colorSchemeDict[key].displayName for key in colorSchemeDict ]
		self.modifierList    = colorschemeGroup.addLabeledControl(modifierBoxLabel, nvdaControls.CustomCheckListBox,
													  choices=self.modifierChoices )



		checkedItems = []
		log.debug(f"USA: {self.modifierChoices}")
		if conf.get("selectedDefault") == 'True':
			checkedItems.append(self.modifierChoices.index(colorSchemeDict["default"].displayName))
		if conf.get("selectedInverted") == 'True':
			checkedItems.append(self.modifierChoices.index(colorSchemeDict["inverted"].displayName))
		if conf.get("selectedGray") == 'True':
			checkedItems.append(self.modifierChoices.index(colorSchemeDict["gray"].displayName))
		if conf.get("selectedHighContrast") == 'True':
			checkedItems.append(self.modifierChoices.index(colorSchemeDict["highContrast"].displayName))
		if conf.get("selectedInvertedHighContrast") == 'True':
			checkedItems.append(self.modifierChoices.index(colorSchemeDict["invertedHighContrast"].displayName))
		if conf.get("selectedBlueLight") == 'True':
			checkedItems.append(self.modifierChoices.index(colorSchemeDict["blueLight"].displayName))
		if conf.get("selectedCustom") == 'True':
			checkedItems.append(self.modifierChoices.index(colorSchemeDict["custom"].displayName))


		self.modifierList.CheckedItems = checkedItems
		self.modifierList.Select(0)


		# # Translators: Button
		self.colorButton = colorschemeGroup.addItem(wx.Button(colorschemeBox, label=_("Manage custom color sch&eme...")))
		self.colorButton.Bind(wx.EVT_BUTTON, self.onColor)
		self.colorButton.Enable(bool(settings.supportedSettings))

		# button for fg depending color schemes
		self.fgColorButton = colorschemeGroup.addItem(wx.Button(colorschemeBox, label=_("Manage App-depending color schemes")))
		self.fgColorButton.Bind(wx.EVT_BUTTON, self.onFGColor)
		self.fgColorButton.Enable(bool(settings.supportedSettings))


		self.Layout()
		return res

	def _enableToggle(self, evt):
		super()._enableToggle(evt)
		settings = self._providerSettings.getSettings()
		self.colorButton.Enable(bool(settings.supportedSettings))

	def onLicense(self, evt):
		# get the license dialog from the global plugin, since we also need it there.
		allPlugins = list(globalPluginHandler.runningPlugins)
		usaPlugin = None
		for ii in allPlugins:
			try:
				# just call a function that we know
				enabled = ii.isUSAEnabled()
				usaPlugin = ii
			except:
				pass
		if usaPlugin:
			USALicenseDialog = usaPlugin.getLicenseDialog()
			with USALicenseDialog(self) as dlg:
				dlg.ShowModal()

	def onColor(self, evt):
		settings = self._providerSettings.getSettings()
		with USACustomColorSchemeDialog(self, settings) as dlg:
			dlg.ShowModal()

	def onFGColor(self, evt):
		settings = self._providerSettings.getSettings()
		with USAFGColorSchemeDialog(self, settings) as dlg:
			dlg.ShowModal()

	def onSave(self):
		# save colorscheme selection
		settings = self._providerSettings.getSettings()
		colorSchemeDict = settings._get_allColorschemes()
		conf = settings.conf
		conf["selectedDefault"] = str( self.modifierList.IsChecked(self.modifierChoices.index(colorSchemeDict["default"].displayName)) )
		conf["selectedInverted"] = str( self.modifierList.IsChecked(self.modifierChoices.index(colorSchemeDict["inverted"].displayName)) )
		conf["selectedGray"] = str( self.modifierList.IsChecked(self.modifierChoices.index(colorSchemeDict["gray"].displayName)) )
		conf["selectedHighContrast"] = str( self.modifierList.IsChecked(self.modifierChoices.index(colorSchemeDict["highContrast"].displayName)) )
		conf["selectedInvertedHighContrast"] = str( self.modifierList.IsChecked(self.modifierChoices.index(colorSchemeDict["invertedHighContrast"].displayName)) )
		conf["selectedCustom"] = str( self.modifierList.IsChecked(self.modifierChoices.index(colorSchemeDict["custom"].displayName)) )
		conf["selectedBlueLight"] = str( self.modifierList.IsChecked(self.modifierChoices.index(colorSchemeDict["blueLight"].displayName)) )
		settings.objectReaderMode = self.objectReaderMode.GetValue()
		conf["caretThickness"] = self.thicknessCtrl.Value
		# conf["caretColor"] = self.colorList.Value[0]

		if self.isRemote.GetValue() and not settings.remoteHandler:
			ui.messageBox(_("Unicorn NVDA addon cannot be found! USA cannot run in remote if Unicorn is not active!"))
			settings.isRemote = False
		else:
			settings.isRemote = self.isRemote.GetValue()


		#

		super().onSave()
		# if update and update.updateChecker and update.updateChecker.IsRunning(): update.updateChecker.Stop()
		# conf = self._providerSettings.getSettings().conf
		# conf["autoUpdateCheck"] = self.autoUpdateCheckbox.Value
		# conf["updateCheckTimeInterval"] = self.updateInterval.Value
		# if not update:
		# 	return
		# if not self.updateInterval.Value:
		# 	conf["updateCheckTime"] = 0
		# 	update.updateChecker = None
		# else:
		# 	update.updateChecker = wx.PyTimer(update.autoUpdateCheck)
		# 	update.updateChecker.Start(update.addonUpdateCheckInterval * 1000, True)
	#
	# def onUpdateCheck(self, evt):
	# 	self.onSave()
	# 	update.progressDialog = gui.IndeterminateProgressDialog(gui.mainFrame,
	# 	# Translators: The title of the dialog presented while checking for add-on updates.
	# 	_("Add-on update check"),
	# 	# Translators: The message displayed while checking for newer version of WinTenApps add-on.
	# 	_("Checking for new version of Ultimate Screen Access add-on..."))
	# 	threading.Thread(target=update.addonUpdateCheck).start()

class USACustomColorSchemeDialog(gui.settingsDialogs.SettingsDialog):

	def makeSettings(self, settingsSizer):
		main_sizer_helper = gui.guiHelper.BoxSizerHelper(self, sizer=settingsSizer)
		conf = self.settings.conf
		self.initialScheme = conf['customColorScheme'].dict()
		self.scheme = conf['customColorScheme']

		# # Translators: Button
		self.matrixButton = main_sizer_helper.addItem(wx.Button(self, label=_("Advanced Settings")))
		self.matrixButton.Bind(wx.EVT_BUTTON, self.onMatrix)


		defaultValue = 100
		brightnessLabelText = _("Brightness (%)")
		self.BrightnessCtrl = main_sizer_helper.addLabeledControl(
			brightnessLabelText,
			nvdaControls.SelectOnFocusSpinCtrl,
			min = 0,
			max = 200,
			initial = int(conf.get('Brightness')) if conf.get('Brightness') else defaultValue,
		)
		self.BrightnessCtrl.Bind(wx.EVT_SPINCTRL, self.onSpinCtrlChange)

		contrastLabelText = _("Contrast (%)")
		self.ContrastCtrl = main_sizer_helper.addLabeledControl(
			contrastLabelText,
			nvdaControls.SelectOnFocusSpinCtrl,
			min = 0,
			max = 200,
			initial = int(conf.get('Contrast')) if conf.get('Contrast') else defaultValue,
		)
		self.ContrastCtrl.Bind(wx.EVT_SPINCTRL, self.onSpinCtrlChange)

		saturationLabelText = _("Saturation (%)")
		self.SaturationCtrl = main_sizer_helper.addLabeledControl(
			saturationLabelText,
			nvdaControls.SelectOnFocusSpinCtrl,
			min = 0,
			max = 500,
			initial = int(conf.get('Saturation')) if conf.get('Saturation') else defaultValue,
		)
		self.SaturationCtrl.Bind(wx.EVT_SPINCTRL, self.onSpinCtrlChange)

		self.HueScaling = 360
		HueRotationLabelText = _("Hue Rotation (degrees)")
		self.HueRotationCtrl = main_sizer_helper.addLabeledControl(
			HueRotationLabelText,
			nvdaControls.SelectOnFocusSpinCtrl,
			min=0,
			max=self.HueScaling,
			initial= int(conf.get('HueRotation')) if conf.get('HueRotation') else 0,
		)
		self.HueRotationCtrl.Bind(wx.EVT_SPINCTRL, self.onSpinCtrlChange)

		blueLightLabelText = _("Blue Light (%)")
		self.BlueLightCtrl = main_sizer_helper.addLabeledControl(
			blueLightLabelText,
			nvdaControls.SelectOnFocusSpinCtrl,
			min = 0,
			max = 150,
			initial = int(conf.get('BlueLight')) if conf.get('BlueLight') else 0,
		)
		self.BlueLightCtrl.Bind(wx.EVT_SPINCTRL, self.onSpinCtrlChange)

		self.checkboxInverse = main_sizer_helper.addItem((wx.CheckBox(self, label="Inverse colours?")))
		self.checkboxInverse.SetValue( conf.get('inverted') == 'True' )
		self.checkboxInverse.Bind(wx.EVT_CHECKBOX, self.onSpinCtrlChange)

		self.checkboxOffset = main_sizer_helper.addItem((wx.CheckBox(self, label="Add offset colour?")))
		self.checkboxOffset.SetValue( conf.get('offset') == 'True' )
		self.checkboxOffset.Bind(wx.EVT_CHECKBOX, self.onSpinCtrlChange)

		self.colorWheel = main_sizer_helper.addItem(wx.ColourPickerCtrl(self, name='colorwheel'))

		self.colorWheel.Bind(wx.EVT_COLOURPICKER_CHANGED, self.onSpinCtrlChange)
		if conf.get('color'):
			if type(conf.get('color')) == tuple:
				colorTuple = conf.get('color')
			else:
				colorTuple = [eval(ii) for ii in conf.get('color')]
		else:
			colorTuple = (0,0,0,0)

		self.colorWheel.SetColour( wx.Colour( colorTuple[0], colorTuple[1], colorTuple[2], colorTuple[3] )  )

	def __init__(self, parent, settings):
		self.settings = settings
		self.rowToConfKeyMap = ("r", "g", "b", "a", "t")
		super().__init__(parent, multiInstanceAllowed=True)

	def hsv_to_rgb(self, h):
		v = 1
		s = 1
		i = int(h * 6.)  # XXX assume int() truncates!
		f = (h * 6.) - i
		p, q, t = int(255 * (v * (1. - s))), int(255 * (v * (1. - s * f))), int(255 * (v * (1. - s * (1. - f))))
		v *= 255
		i %= 6
		if i == 0: return (v, t, p)
		if i == 1: return (q, v, p)
		if i == 2: return (p, v, t)
		if i == 3: return (p, q, v)
		if i == 4: return (t, p, v)
		if i == 5: return (v, p, q)

	def getBSCMatrix(self):
		# from https://docs.rainmeter.net/tips/colormatrix-guide/
		# https://flylib.com/books/en/2.72.1/matrix_operations_in_image_processing.html
		# https://www.c-sharpcorner.com/uploadfile/mahesh/color-mapping-using-color-objects-in-gdi/
		B = ( self.BrightnessCtrl.GetValue() - 100.) / 100.
		C = self.ContrastCtrl.GetValue() / 100.
		S = self.SaturationCtrl.GetValue() / 100.
		rgbOffset = self.colorWheel.GetColour().Get()
		h = self.HueRotationCtrl.GetValue()/self.HueScaling
		addOffsetColor = self.checkboxOffset.GetValue()
		invert = self.checkboxInverse.GetValue()
		blueLight = self.BlueLightCtrl.GetValue() / 100.
		sumRGB = (rgbOffset[0] + rgbOffset[1] + rgbOffset[2])/3.0
		rOffset = (rgbOffset[0] - sumRGB) / 255.0 if addOffsetColor else 0.0
		gOffset = (rgbOffset[1] - sumRGB) / 255.0 if addOffsetColor else 0.0
		bOffset = (rgbOffset[2] - sumRGB) / 255.0 if addOffsetColor else 0.0

		t = ( 1.0 - C )/2.0

		lumR = 0.3086
		lumG = 0.6094
		lumB = 0.0820
		sr = (1.0 - S)*lumR
		sg = (1.0 - S)*lumG
		sb = (1.0 - S)*lumB

		RGBmats = [
			[C * (sr + S)- (blueLight * 0.2), C * sr, C * sr],
			[C * sg, C * (sg + S) - (blueLight * 0.5), C * sg],
			[C * sb, C * sb, C * (sb + S) - blueLight]
		]

		RGBmat = mo.rotateHue(2 * mo.pi * h)
		RGBmat = mo.matrix_multiply(RGBmat, RGBmats)
		rowR = [RGBmat[0][0], RGBmat[0][1], RGBmat[0][2], 0, 0]
		rowG = [RGBmat[1][0], RGBmat[1][1], RGBmat[1][2], 0, 0]
		rowB = [RGBmat[2][0], RGBmat[2][1], RGBmat[2][2], 0, 0]
		rowA = [0, 0, 0, 1, 0]
		rowW = [t + B + rOffset, t + B + gOffset, t + B + bOffset, 0, 1]
		BSCMatrix = [rowR, rowG, rowB, rowA, rowW]

		if invert:
			BSCMatrix = mo.matrix_multiply(mo.inverseMatrix, BSCMatrix)

		log.debug(f"USA: {RGBmats} \n {BSCMatrix}")
		return BSCMatrix

	def onSpinCtrlChange(self, evt):
		BCSMatrix = self.getBSCMatrix()
		for r in range(0, 5):
			for c in range(0, 5):
				self.scheme[self.rowToConfKeyMap[r]][c] = BCSMatrix[r][c]
		self.settings.colorScheme = self.settings.colorScheme
		evt.Skip()

	def onMatrix(self, evt):
		with USACustomColorSchemeMatrix(self, self.settings) as dlg:
			dlg.ShowModal()

		self.initialScheme = self.settings.conf['customColorScheme'].dict()
		self.onCancel(wx.EVT_BUTTON)

	def onOk(self, evt):
		self.settings.conf['Brightness'] = self.BrightnessCtrl.GetValue()
		self.settings.conf['Contrast'] = self.ContrastCtrl.GetValue()
		self.settings.conf['Saturation'] = self.SaturationCtrl.GetValue()
		self.settings.conf['HueRotation'] = self.HueRotationCtrl.GetValue()
		self.settings.conf['inverted'] = str( self.checkboxInverse.GetValue() )
		self.settings.conf['offset'] = str( self.checkboxOffset.GetValue() )
		self.settings.conf['BlueLight'] = self.BlueLightCtrl.GetValue()

		BCSMatrix = self.getBSCMatrix()
		for r in range(0, 5):
			for c in range(0, 5):
				self.settings.conf['customColorScheme'][self.rowToConfKeyMap[r]][c]= BCSMatrix[r][c]
		self.settings.colorScheme = self.settings.colorScheme

		if self.settings.conf['offset']:
			self.settings.conf['color'] = self.colorWheel.GetColour().Get(includeAlpha = True)
		super().onOk(evt)

	def onCancel(self, evt):
		for r in range(0, 5):
			for c in range(0, 5):
				self.settings.conf['customColorScheme'][self.rowToConfKeyMap[r]][c] = self.initialScheme[self.rowToConfKeyMap[r]][c]
		self.settings.colorScheme = self.settings.colorScheme

		super().onCancel(evt)

class USACustomColorSchemeMatrix(wx.Dialog):

	def __init__(self, parent, settings):
		super().__init__(parent, id=wx.ID_ANY, title=_("Change color scheme"))
		self.settings = settings
		conf = settings.conf
		self.initialScheme = conf['customColorScheme'].dict()
		self.scheme = conf['customColorScheme']
		main_sizer = wx.BoxSizer(wx.VERTICAL)
		main_sizer_helper = gui.guiHelper.BoxSizerHelper(self, sizer=main_sizer)
		self.rowToConfKeyMap = ("r", "g", "b", "a", "t")
		gridSizer=wx.GridBagSizer(
			hgap=guiHelper.SPACE_BETWEEN_BUTTONS_HORIZONTAL,
			vgap=guiHelper.SPACE_BETWEEN_BUTTONS_VERTICAL
		)
		gridSizer.Rows=6
		gridSizer.Cols=6
		self.spinControls = []
		for r in range(gridSizer.Rows):
			if r != 0:
				l = []
				self.spinControls.append(l)
			for c in range(gridSizer.Cols):
				if r == 0 and c == 0:
					continue
				if r == 0 and c != 0:
					label = wx.StaticText(self, label=self.rowToConfKeyMap[c-1])
					gridSizer.Add(label, pos=(r, c))
					continue
				if c == 0 and r != 0:
					label = wx.StaticText(self, label=self.rowToConfKeyMap[r-1])
					gridSizer.Add(label, pos=(r, c))
					continue
				control = wx.SpinCtrlDouble(
					self,
					initial=0.0,
					inc=0.1,
					min=-10.0,
					max=10.0
				)
				control.Value = self.scheme[self.rowToConfKeyMap[r - 1]][c - 1]
				control.SetToolTip(f"{self.rowToConfKeyMap[r-1]}:{self.rowToConfKeyMap[c-1]}")
				control.Bind(wx.EVT_SPINCTRLDOUBLE, self.onSpinCtrlChange, source=control)
				gridSizer.Add(control, pos=(r, c))
				l.append(control)
		main_sizer.Add(gridSizer, border = gui.guiHelper.BORDER_FOR_DIALOGS, flag=wx.ALL)
		main_sizer_helper.addDialogDismissButtons(self.CreateButtonSizer(wx.OK | wx.CANCEL))
		self.Bind(wx.EVT_BUTTON, self.onOk, id=wx.ID_OK)
		self.Bind(wx.EVT_BUTTON, self.onCancel, id=wx.ID_CANCEL)
		main_sizer.Fit(self)
		self.SetSizer(main_sizer)
		self.Center(wx.BOTH)
		self.Show()

	def onSpinCtrlChange(self, evt):
		obj = evt.EventObject
		for r, lst in enumerate(self.spinControls):
			for c, control in enumerate(lst):
				if obj is control:
					self.scheme[self.rowToConfKeyMap[r]][c] = evt.Value
					break
			else:
				continue
			break
		else:
			log.error(f"{obj!r} not known to handler")
			return
		self.settings.colorScheme = self.settings.colorScheme
		evt.Skip()

	def onOk(self, evt):
		for r, lst in enumerate(self.spinControls):
			for c, control in enumerate(lst):
				self.settings.conf['customColorScheme'][self.rowToConfKeyMap[r]][c] = control.Value
		self.settings.colorScheme = self.settings.colorScheme
		evt.Skip()

	def onCancel(self, evt):
		for r, lst in enumerate(self.spinControls):
			key = self.rowToConfKeyMap[r]
			for c, control in enumerate(lst):
				self.settings.conf['customColorScheme'][key][c] = self.initialScheme[key][c]
		self.settings.colorScheme = self.settings.colorScheme
		evt.Skip()

class USAFGColorSchemeDialog(gui.settingsDialogs.SettingsDialog):
	def makeSettings(self, settingsSizer):
		main_sizer_helper = gui.guiHelper.BoxSizerHelper(self, sizer=settingsSizer)
		conf = self.settings.conf
		self.initialScheme = conf['customColorScheme'].dict()
		self.scheme = conf['customColorScheme']

		self.fgColorschemesDict = conf['fgColorschemes']

		self.modifierChoices = [key for key, item in list(self.fgColorschemesDict.items()) if item]
		self.modifierList = main_sizer_helper.addLabeledControl(_('remove foreground depending color schemes:'), nvdaControls.CustomCheckListBox,
															   choices=self.modifierChoices)

	def __init__(self, parent, settings):
		self.settings = settings
		super().__init__(parent, multiInstanceAllowed=True)

	def onOk(self, evt):

		for key, item in list(self.fgColorschemesDict.items()):
			if item:
				if self.modifierList.IsChecked(self.modifierChoices.index(key)):
					self.settings.conf['fgColorschemes'][key] = None


		super().onOk(evt)

	def onCancel(self, evt):

		super().onCancel(evt)
