
from ctypes import *
from ctypes.wintypes import *
import os
import wx # type: ignore
import addonHandler
import gui
import core
import queueHandler
import watchdog
from logHandler import log
import config
import datetime
addonHandler.initTranslation()

a4uLicenseMessageMapping = {
    "to verify there are activations remaining, and the API Key and Product ID are correct.": _("This license key is not valid for this product."), # type: ignore
    "customer account does not exist for this API Key.": _("This license key is not valid."), # type: ignore
    "not activate API Key. No API resources available.": _("No activations remaining.") # type: ignore
}
# something can go wrong with the conversion in the dll when tranlating the json return to some string class in c++,
# this causes that in the beginning of the error message some chinese characters appear, thus we check if the message
# ends with a known message and return the correct message
def getMessageEndWithA4uLicenseMessage(message):
    for key in a4uLicenseMessageMapping:
        if message.endswith(key):
            return a4uLicenseMessageMapping[key]
    return None

# load lib modules
addon: addonHandler.Addon = addonHandler.getCodeAddon()
libUSA = addon.loadModule("lib.libUSA")
licenseExtension = addon.loadModule("lib.licenseExtension")
access4uReturnCodes = addon.loadModule("lib.access4uReturnCodes")
demoHandler = addon.loadModule("lib.demoHandler")

def ActivateLicense(licenseKey):
    (success, message) = libUSA.ActivateLicense(licenseKey)
    return (success, message)

def DeactivateLicense():
    (success, message) = libUSA.DeactivateLicense()
    return (success, message)

class USALicenseDialog(wx.Dialog):

    def restartNVDA(self):
        
        queueHandler.queueFunction(queueHandler.eventQueue, core.restart)

    def checkIfInRemoteMode(self):
        import vision
        allProviders = vision.handler._allProviders
        for provider in allProviders:
            if provider.providerId == 'usa':
                return provider.providerClass.getSettings().isRemote
        return False      

    def addTopMessage(self):
        addonManifest = addonHandler.getCodeAddon().manifest
        currentVersion = addonManifest['version']
        
        # Load logo image
        addon: addonHandler.Addon = addonHandler.getCodeAddon()
        path = addon.path+"\\BabbageIcon.png"
        image = wx.Image(path, wx.BITMAP_TYPE_PNG)
        image = image.Scale(115, 115, wx.IMAGE_QUALITY_HIGH)
        image_ctrl = wx.StaticBitmap(self, wx.ID_ANY, (wx.Bitmap(image)), pos=(0,0))
        self.top_sizer.Add(image_ctrl, 0, wx.LEFT, 15)
      
        if self.isLicensed:
            log.debugWarning("USA is licensed")
            message = _("Your copy of Ultimate Screen Access {currentVersion} is properly licensed.").format(currentVersion=currentVersion)
        else:
            # always show when USA is not licensed
            log.debugWarning("USA not licensed")
            self.shouldShowAtStart = True
            message = _("Your copy of Ultimate Screen Access {currentVersion} doesn't seem to be licensed.").format(currentVersion=currentVersion)
        demoHandler.handler.status = demoHandler.handler.loadFromRegistry()
        log.debugWarning(f"demo status: {demoHandler.handler.status}")
        if len(demoHandler.handler.status) > 0:
            if demoHandler.handler.status[0] == demoHandler.DEMO_FOUND:
                log.debugWarning("found demo")
                self.shouldShowAtStart = True
                version = demoHandler.handler.status[1]
                dbId = demoHandler.handler.status[2]
                timestamp = demoHandler.handler.status[3]
                RemainingTime = datetime.datetime.fromtimestamp(timestamp) - datetime.datetime.now()                
                message =  _("You are currently running a demo version of USA, you have {days} days remaining").format(days=RemainingTime.days)   
                message += "\n"+ _("Contact info@babbage.com or call 0165 536 156 \nfor questions regarding the purchase or use of the product")
            elif demoHandler.handler.status[0] == demoHandler.DEMO_OVER:
                log.debugWarning("found demo but was over")
                self.shouldShowAtStart = True
                self.demoDatabaseID = demoHandler.handler.status[2]
                message = _("Your demo for Ultimate Screen Access has expired.") 
                message +="\n" +  _("You can enter a valid license key, or contact support@babbage.com if this message is unexpected.")
                message +="\n" +  _("The id for your demo is {id}").format(id = self.demoDatabaseID)
                self.demoOver = True
            elif demoHandler.handler.status[0] == demoHandler.DEMO_INACTIVE:
                log.debugWarning("found demo but is inactive")
                self.shouldShowAtStart = True
                self.demoDatabaseID = demoHandler.handler.status[2]
        
        if self.checkIfInRemoteMode():
            log.debugWarning("USA runs in remote mode")
            self.shouldShowAtStart = False
            message = _(f"USA is running in Remote Mode. No license is needed on this machine.")
        
        message = "Babbage\ninfo@babbage.com\n0165 536 156\n\n" + message
        self.top_sizer.Add(wx.StaticText(self, label=message),  wx.SizerFlags(0).Align(wx.LEFT).Border(wx.LEFT, 30))

    def addLicenseKeyOrDeactivateButton(self):
        #Translators: The input field to enter the USA license key
        if not self.isLicensed or (demoHandler.handler.status[0] == demoHandler.DEMO_FOUND or self.demoOver):
            activationButton = self.buttontop_sizer_helper.addItem(wx.Button(self, label=_("activate with key")), flag = wx.EXPAND)
            activationButton.Bind(wx.EVT_BUTTON, self.activate)
        else:
            deActivationButton = self.buttontop_sizer_helper.addItem(wx.Button(self, label=_("deactivate key")), flag = wx.EXPAND)
            deActivationButton.Bind(wx.EVT_BUTTON, self.deactivate)

    def addButtons(self):
        runRemoteMode = self.buttontop_sizer_helper.addItem(wx.Button(self, label=_("turn on USA Remote")), flag = wx.EXPAND)
        runRemoteMode.Bind(wx.EVT_BUTTON, self.runRemoteMode)
        if self.checkIfInRemoteMode(): runRemoteMode.Disable()
        offlineActivation = self.buttonbot_sizer_helper.addItem(wx.Button(self, label=_("Offline activation")), flag = wx.EXPAND)
        offlineActivation.Bind(wx.EVT_BUTTON, self.offlineActivation)
        requestDemo = self.buttonbot_sizer_helper.addItem(wx.Button(self, label=_("Request demo")), flag = wx.EXPAND)
        requestDemo.Bind(wx.EVT_BUTTON, self.requestDemo)
        # isLicensed is true when the licensed still needs to be determined. When called at the start, if a demo is found
        # isLicensed still returns true in the first few seconds. This is why we check for demoOver as well
        if self.checkIfInRemoteMode() or (self.isLicensed and not self.demoOver) or demoHandler.handler.status[0] == demoHandler.DEMO_FOUND: 
            offlineActivation.Disable()
            #autoActivation.Disable()
            requestDemo.Disable()
        
        # Resize these buttons to have the same size as buttontop.GetItem(2), which is the shortest. 
        # Add the buttonlist to the mainsizer list
        self.buttontop_sizer_helper.sizer.GetItem(0).SetMinSize(self.buttontop_sizer_helper.sizer.GetItem(2).Size)
        self.buttonbot_sizer_helper.sizer.GetItem(0).SetMinSize(self.buttontop_sizer_helper.sizer.GetItem(2).Size)
        self.buttonbot_sizer_helper.sizer.GetItem(2).SetMinSize(self.buttontop_sizer_helper.sizer.GetItem(2).Size)
        self.main_sizer_helper.addItem(self.buttontop_sizer_helper, border = (1), flag=wx.ALL)
        self.main_sizer_helper.addItem(self.buttonbot_sizer_helper, border = (1), flag=wx.ALL)
        
        bHelper = self.main_sizer_helper.addDialogDismissButtons(gui.guiHelper.ButtonHelper(wx.HORIZONTAL))
        closeButton = bHelper.addButton(self, wx.ID_CLOSE, label=_("&Close"))
        closeButton.Bind(wx.EVT_BUTTON, self.onClose)
        closeButton.SetFocus()


    def __init__(self, parent):
        # Translators: Title of the license dialog
        windowStyle = wx.DEFAULT_DIALOG_STYLE | wx.CENTER
        super().__init__(parent, id=wx.ID_ANY, title=_("Manage USA License"), style = windowStyle)
        self.isLicensed = libUSA.IsLicensed()
        self.isAutomaticActivation = False
        self.isOfflineActivation = False
        self.isDemo = False
        self.demoOver = False
        self.shouldShowAtStart = False
        self.demoDatabaseID = 0
        

        self.main_sizer = wx.BoxSizer(wx.VERTICAL)
        self.main_sizer_helper = gui.guiHelper.BoxSizerHelper(self, orientation=wx.VERTICAL)
        self.top_sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.buttontop_sizer_helper = gui.guiHelper.BoxSizerHelper(self, orientation=wx.HORIZONTAL)
        self.buttonbot_sizer_helper = gui.guiHelper.BoxSizerHelper(self, orientation=wx.HORIZONTAL)
        self.addTopMessage()
        self.main_sizer.Add(self.top_sizer, 0, wx.RIGHT, 10)
        self.addLicenseKeyOrDeactivateButton()
        self.addButtons()        

        self.main_sizer.Add(self.main_sizer_helper.sizer, border = gui.guiHelper.BORDER_FOR_DIALOGS, flag=wx.ALL)
        self.main_sizer.Fit(self)
        self.SetSizer(self.main_sizer)
        self.Center(wx.BOTH )

    def activate(self, evt):
        success = False
        with activationDialog(self) as dlg:
            dlg.ShowModal()
            success = dlg.success
        if success:
            self.Close()
            self.Destroy()  
            wx.CallAfter(self.restartNVDA)

    def deactivate(self, evt):
        try:
            success, message = watchdog.cancellableExecute(DeactivateLicense)
        except Exception as e:
            success = False
            message = _(f"There was a timeout while performing your request. {e}")
            log.error("Activation error", exc_info=True)
        if not success:
            if getMessageEndWithA4uLicenseMessage(message):
                message = getMessageEndWithA4uLicenseMessage(message)
            gui.messageBox(_("An error has occured:\n{error}").format(error=message),
                    _("Error"), wx.OK | wx.ICON_ERROR)
        else:
            gui.messageBox(_("Ultimate Screen Access has been deactivated! NVDA will be restarted. \nAdditional info: {message}").format(message=message),
                    _("Congratulations!"), wx.OK | wx.ICON_EXCLAMATION)
        if success:
            self.Close()
            self.Destroy()
            wx.CallAfter(self.restartNVDA)

    def offlineActivation(self, evt):

        with offlineDialog(self) as dlg:
            dlg.ShowModal()

        self.Close()
        self.Destroy()

    def automaticActivation(self, evt):

        with automaticActivationDialog(self) as dlg:
            dlg.ShowModal()

        self.Close()
        self.Destroy()

    def requestDemo(self, evt):

        with requestDemoDialog(self, self.demoDatabaseID) as dlg:
            dlg.ShowModal()
            success = dlg.success
        if success:
            self.Close()
            self.Destroy()  

    def runRemoteMode(self, evt):

        with remoteModeDialog(self) as dlg:
            dlg.ShowModal()
            success = dlg.success
        if success:
            self.Close()
            self.Destroy()  
    
    def onClose(self, evt):
        self.Close()

class remoteModeDialog(wx.Dialog):

    def __init__(self, parent):
        windowStyle = wx.DEFAULT_DIALOG_STYLE | wx.STAY_ON_TOP | wx.CENTER
        super().__init__(parent, id=wx.ID_ANY, title=_("remote mode"), style = windowStyle)
        
        self.main_sizer = wx.BoxSizer(wx.VERTICAL)
        self.main_sizer_helper = gui.guiHelper.BoxSizerHelper(self, orientation=wx.VERTICAL)
        message = _("Remote Mode is to connect the USA addon on a RDS to a USA addon on the local computer.") +"\n"
        message += _("Only turn on this setting when this USA addon is running in a RDS.")
        self.main_sizer.Add(wx.StaticText(self, label=message),  wx.SizerFlags(0).Align(wx.LEFT).Border(wx.ALL, 5))

        self.Bind(wx.EVT_BUTTON, self.on_ok, id=wx.ID_OK)
        self.main_sizer.Add(self.main_sizer_helper.sizer, border = gui.guiHelper.BORDER_FOR_DIALOGS, flag=wx.ALL)
        self.main_sizer_helper.addDialogDismissButtons(self.CreateButtonSizer( wx.OK| wx.CANCEL))
        self.main_sizer.Fit(self)
        self.SetSizer(self.main_sizer)
        self.Center(wx.BOTH )
        self.success = False

    def on_ok(self, evt):
        self.remoteMode()
        evt.Skip()
    
    def remoteMode(self):
        self.success = False
        try:
            import vision
            allProviders = vision.handler._allProviders
            for provider in allProviders:
                if provider.providerId == 'usa':
                    provider.providerClass.getSettings().isRemote = True
                    self.success = True           
        except Exception as e:
            self.success = False
            message = _("There was an error while setting remote mode for USA. \n{e}").format(e)
            log.error(message, exc_info=True)
                          
        if not self.success:
            gui.messageBox(_("Setting USA to remote mode has failed."),
                    _("Error"), wx.OK | wx.ICON_ERROR)
        else:
            gui.messageBox(_("USA is set to run in remote mode! Please restart NVDA."),
                    _("success"), wx.OK | wx.ICON_EXCLAMATION)
    
        if self.success:
            self.Close()

class requestDemoDialog(wx.Dialog):

    def __init__(self, parent, demo_databaseID = 0):
        windowStyle = wx.DEFAULT_DIALOG_STYLE | wx.STAY_ON_TOP | wx.CENTER
        super().__init__(parent, id=wx.ID_ANY, title=_("Request Demo"), style = windowStyle)
        
        self.main_sizer = wx.BoxSizer(wx.VERTICAL)
        self.main_sizer_helper = gui.guiHelper.BoxSizerHelper(self, orientation=wx.VERTICAL)
        message = _("A demo is 1 month valid and tied to this computer.")
        message += "\n" + _("To request a demo, an internet connection is required.")
        message += "\n" + _("Would you like to request a demo?")
        if demo_databaseID:
            message = _("If you want to extend your demo period, please contact Babbage at support@babbage.com. Continue with the request if this is done.") 
            message += _("\n A previous demo is found on this system with id {demo_databaseID}.").format(demo_databaseID=demo_databaseID)
            
            
        self.main_sizer.Add(wx.StaticText(self, label=message),  wx.SizerFlags(0).Align(wx.LEFT).Border(wx.ALL, 5))

        self.Bind(wx.EVT_BUTTON, self.on_ok, id=wx.ID_OK)
        self.main_sizer.Add(self.main_sizer_helper.sizer, border = gui.guiHelper.BORDER_FOR_DIALOGS, flag=wx.ALL)
        self.main_sizer_helper.addDialogDismissButtons(self.CreateButtonSizer( wx.OK| wx.CANCEL))
        self.main_sizer.Fit(self)
        self.SetSizer(self.main_sizer)
        self.Center(wx.BOTH )
        self.success = False

    def on_ok(self, evt):
        wx.CallAfter(self.requestDemo)
    
    def requestDemo(self):
        returnCode = access4uReturnCodes.a4u_fail
        try:
            regPath = os.path.join("SOFTWARE", "ACCESS4U", demoHandler.usaVersions[0])
            returnCode = watchdog.cancellableExecute(licenseExtension.requestDemo, regPath, demoHandler.productID, demoHandler.usaVersions[0])
            log.debugWarning(f"got returncode {returnCode}")
        except Exception as e:
            returnCode = access4uReturnCodes.a4u_unknown
            message = _("There was an error while performing your request. \n{e}").format(e=e)
            log.error(_("Activation error"), exc_info=True)
            gui.messageBox(message,
                    _("Error"), wx.OK | wx.ICON_ERROR)
            return
        
        self.success = returnCode == access4uReturnCodes.a4u_success
        if self.success:
            gui.messageBox(_("Ultimate Screen Access DEMO has been activated! NVDA will be restarted."),
                    _("Congratulations!"), wx.OK | wx.ICON_EXCLAMATION)
            self.Close()
            queueHandler.queueFunction(queueHandler.eventQueue, core.restart)
            return
        elif returnCode == access4uReturnCodes.a4u_unknown:
            message = _("Something went wrong when trying to request a demo.")
            message += "\n" + _("Contact support@babbage.com if this message is unexpected.")
            gui.messageBox(message,
                _("Error"), wx.OK | wx.ICON_ERROR)
            self.Close()
            return
        message = _("The demo request was denied by the server.")
        message += "\n" + _("Contact support@babbage.com if this message is unexpected.")
        gui.messageBox(message,
                _("Error"), wx.OK | wx.ICON_ERROR)
        self.Close()

class activationDialog(wx.Dialog):

    def __init__(self, parent):
        windowStyle = wx.DEFAULT_DIALOG_STYLE | wx.STAY_ON_TOP | wx.CENTER
        super().__init__(parent, id=wx.ID_ANY, title=_("Manage USA License"), style = windowStyle)
        
        self.main_sizer = wx.BoxSizer(wx.VERTICAL)
        self.main_sizer_helper = gui.guiHelper.BoxSizerHelper(self, orientation=wx.VERTICAL)

        self.key = self.main_sizer_helper.addLabeledControl(_("&License Key:"), wx.TextCtrl)

        self.Bind(wx.EVT_BUTTON, self.on_ok, id=wx.ID_OK)
        self.main_sizer.Add(self.main_sizer_helper.sizer, border = gui.guiHelper.BORDER_FOR_DIALOGS, flag=wx.ALL)
        self.main_sizer_helper.addDialogDismissButtons(self.CreateButtonSizer( wx.OK| wx.CANCEL))
        self.main_sizer.Fit(self)
        self.SetSizer(self.main_sizer)
        self.Center(wx.BOTH )
        self.key.SetFocus()

    def on_ok(self, evt):
        self.activate()
    
    def activate(self):
        if not self.key.Value:
            gui.messageBox(_("You must enter a valid license key."), _("Error"), wx.OK | wx.ICON_ERROR)
            self.key.SetFocus()
            return

        try:
            self.success, message = watchdog.cancellableExecute(ActivateLicense, self.key.Value)
        except Exception as e:
            self.success = False
            message = _("There was an error while performing your request. \n{e}").format(e)
            log.error(_("Activation error"), exc_info=True)


        if not self.success:
            if getMessageEndWithA4uLicenseMessage(message):
                message = getMessageEndWithA4uLicenseMessage(message)
            gui.messageBox(_("An error has occured:\n{error}").format(error=message),
                    _("Error"), wx.OK | wx.ICON_ERROR)
        else:
            gui.messageBox(_("Ultimate Screen Access has been activated! NVDA will be restarted. \nAdditional info: {message}").format(message=message),
                    _("Congratulations!"), wx.OK | wx.ICON_EXCLAMATION)
            queueHandler.queueFunction(queueHandler.eventQueue, core.restart)
        if self.success:
            self.Close()

class automaticActivationDialog(wx.Dialog):

    def __init__(self, parent):
        # Translators: Title of the license dialog
        super().__init__(parent, id=wx.ID_ANY, title=_("automatic Activation"))
        self.isLicensed = libUSA.IsLicensed()
        addonManifest = addonHandler.getCodeAddon().manifest
        currentVersion = addonManifest['version']

        self.autoActivationEnabled = config.conf.get("usaAutoActivation", False)
        self.licenseForAutoActivation = config.conf.get("usaAutoActivationLicenseKey", "")

        message = _("We offer automatic activations for non-presistent enviroments. With this USA communicates with the license server whenever it starts up.")

        main_sizer = wx.BoxSizer(wx.VERTICAL)
        main_sizer.Add(wx.StaticText(self, label=message))
        main_sizer_helper = gui.guiHelper.BoxSizerHelper(self, orientation=wx.VERTICAL)

        productChoices = ["USA 2024", "USA 2025"]
        self.productIDs = [ 1356, 1357]
        self.regPath = ["USA2024", "USA2025"]
        self.productID = main_sizer_helper.addLabeledControl("&license version", wx.Choice, choices=productChoices)
        self.productID.SetSelection(0)

        self.lic = main_sizer_helper.addLabeledControl(_("&license key:"), wx.TextCtrl)
        self.lic.Enabled = True
        self.lic.Value = self.licenseForAutoActivation

        self.automaticActivation = main_sizer_helper.addItem(wx.Button(self, label=_("enable automatic activation")))
        self.automaticActivation.Bind(wx.EVT_BUTTON, self.automaticActivation_Click)

        if self.autoActivationEnabled:
            self.automaticActivation.SetLabel(_("disable automatic activation"))
        else:
            self.automaticActivation.SetLabel(_("enable automatic activation"))

        main_sizer_helper.addDialogDismissButtons(self.CreateButtonSizer(wx.OK | wx.CANCEL))
        self.Bind(wx.EVT_BUTTON, self.on_ok, id=wx.ID_OK)
        main_sizer.Add(main_sizer_helper.sizer, border = gui.guiHelper.BORDER_FOR_DIALOGS, flag=wx.ALL)
        main_sizer.Fit(self)
        self.SetSizer(main_sizer)
        self.Center(wx.BOTH )
        self.Show()

    def automaticActivation_Click(self, evt):

        if not self.autoActivationEnabled and not len(self.lic.Value.strip()) > 0:
            gui.messageBox(_("No license key filled in!"))
            return


        self.autoActivationEnabled = not self.autoActivationEnabled
        if self.autoActivationEnabled:
            self.automaticActivation.SetLabel(_("disable automatic activation"))
            productID = self.productIDs[self.productID.GetSelection()]
            license = self.lic.Value
            config.conf['usaAutoActivationProductID'] = productID
            config.conf['usaAutoActivationLicenseKey'] = license
            gui.messageBox("automatic activation enabled")



        else:
            self.automaticActivation.SetLabel(_("enable automatic activation"))
            gui.messageBox(_("automatic activation disabled"))

        evt.Skip()

    def on_ok(self, evt):
        config.conf['usaAutoActivation'] = self.autoActivationEnabled
        self.Close()
        self.Destroy()

class offlineDialog(wx.Dialog):

    def __init__(self, parent):
        # Translators: Title of the license dialog
        super().__init__(parent, id=wx.ID_ANY, title=_("offline Activation"))
        self.isLicensed = libUSA.IsLicensed()
        addonManifest = addonHandler.getCodeAddon().manifest
        currentVersion = addonManifest['version']

        message = _("We offer offline activations as well. For this you need to mail your licensekey and instance ID to support@babbage.com. \nWe will send you an unique activation hash so that USA is licensed without the need of verifying online.")

        main_sizer = wx.BoxSizer(wx.VERTICAL)
        main_sizer.Add(wx.StaticText(self, label=message))
        main_sizer_helper = gui.guiHelper.BoxSizerHelper(self, orientation=wx.VERTICAL)
        #Translators: The input field to enter the USA license key


        productChoices = ["USA 2024", "USA 2025", "USA 2026", "USA 2027"]
        self.productIDs = [1356, 1357, 1752, 2010]
        self.regPath = ["USA2024", "USA2025", "USA2026", "USA2027"]
        self.productID = main_sizer_helper.addLabeledControl("&license version", wx.Choice, choices=productChoices)
        self.productID.SetSelection(0)
        self.productID.Bind(wx.EVT_CHOICE, self.changedProduct)

        id = licenseExtension.getInstanceID(self.productIDs[self.productID.GetSelection()])

        self.key = main_sizer_helper.addLabeledControl(_("&Instance ID:"), wx.TextCtrl)
        self.key.Enabled = True
        self.key.Value = id
        self.lic = main_sizer_helper.addLabeledControl(_("&license key:"), wx.TextCtrl)
        self.lic.Enabled = True
        self.activationHash = main_sizer_helper.addLabeledControl(_("&Activation hash:"), wx.TextCtrl)
        self.activationHash.Enabled = True

        main_sizer_helper.addDialogDismissButtons(self.CreateButtonSizer(wx.OK | wx.CANCEL))
        self.Bind(wx.EVT_BUTTON, self.on_ok, id=wx.ID_OK)
        main_sizer.Add(main_sizer_helper.sizer, border = gui.guiHelper.BORDER_FOR_DIALOGS, flag=wx.ALL)
        main_sizer.Fit(self)
        self.SetSizer(main_sizer)
        self.Center(wx.BOTH )
        self.Show()
        self.key.SetFocus()


    def changedProduct(self, evt):
        id = licenseExtension.getInstanceID(self.productIDs[self.productID.GetSelection()])
        self.key.Value = id
        evt.Skip()
    def on_ok(self, evt):
        if len(self.activationHash.Value.split()) > 0:

            regPath = os.path.join("SOFTWARE", "ACCESS4U",self.regPath[self.productID.GetSelection()])
            licenseExtension.saveDataToRegistry(self.lic.Value, self.activationHash.Value, regPath)


        self.Close()
        self.Destroy()





