"""
Babbage B.V., 2019-2023

visionEnhancementProviders/_base/libUSA.py
Used to interact with the libUSA.dll
"""

from ctypes import *
from ctypes.wintypes import *
import os
from logHandler import log
from . import access4uReturnCodes
log.info("importing libUSA")
class MAGCOLOREFFECT(Structure):
	_fields_ = (("transform", c_float*5*5),)

winLicenseError = 7056

MagnifierType = c_int
MAG_TYPE_SCREEN = 0
MAG_TYPE_WINDOW = 1

magWindows = dict()

def errCheck(result, func, args):
	if result >0:
		raise WinError(result)
	return args

try:
	libUSA = windll.libUSA
except:
	libUSA = windll.LoadLibrary(os.path.join(os.path.abspath(os.path.dirname(__file__)), "libUSA.dll"))

# Prototype declarations
InitializeFuncType = WINFUNCTYPE(DWORD, POINTER(HWND), MagnifierType)
InitializeArgTypes = ((2, "pHwnd"),(1,"magnifierType",MAG_TYPE_SCREEN),)
ActivateLicenseFuncType = WINFUNCTYPE(c_wchar_p, c_wchar_p, POINTER(BOOL))
DeactivateLicenseFuncType = WINFUNCTYPE(c_wchar_p, POINTER(BOOL))
TerminateFuncType = WINFUNCTYPE(DWORD)
IsLicensedFuncType = WINFUNCTYPE(BOOL)
CreateMagnifierWindowFuncType = WINFUNCTYPE(DWORD, POINTER(HWND), RECT)
CreateMagnifierWindowArgTypes = ((2, "pHwnd"), (1, "rect"),)
DestroyMagnifierWindowFuncType = WINFUNCTYPE(DWORD, HWND)
DestroyMagnifierWindowArgTypes = ((1, "hwnd"),)
SetTrackingRectangleFuncType = WINFUNCTYPE(DWORD, HWND, RECT, RECT)
SetTrackingRectangleArgTypes = ((1, "hwnd"), (1, "rect"), (1, "boundary", RECT(-1,-1,-1,-1)),)
GetTrackingRectangleFuncType = WINFUNCTYPE(DWORD, HWND, POINTER(RECT))
GetTrackingRectangleArgTypes = ((1, "hwnd"), (2, "rect"),)
SetTrackingOffsetsFuncType = WINFUNCTYPE(DWORD, HWND, POINT)
SetTrackingOffsetsArgTypes = ((1, "hwnd"), (1, "offsets"),)
GetTrackingOffsetsFuncType = WINFUNCTYPE(DWORD, HWND, POINTER(POINT))
GetTrackingOffsetsArgTypes = ((1, "hwnd"), (2, "offsets"),)
SetMagnificationLevelFuncType = WINFUNCTYPE(DWORD, HWND, c_float)
SetMagnificationLevelArgTypes = ((1, "hwnd"), (1, "magLevel"),)
GetMagnificationLevelFuncType = WINFUNCTYPE(DWORD, HWND, POINTER(c_float))
GetMagnificationLevelArgTypes = ((1, "hwnd"),(2, "magLevel"),)
SetMagnificationStepsFuncType = WINFUNCTYPE(DWORD, HWND, c_int)
SetMagnificationStepsArgTypes = ((1, "hwnd"), (1, "MagnificationSteps"))
GetMagnificationStepsFuncType = WINFUNCTYPE(DWORD, HWND, POINTER(c_int))
GetMagnificationStepsArgTypes = ((1, "hwnd"), (2, "MagnificationSteps"))
SetMagnificationStepTimeFuncType = WINFUNCTYPE(DWORD, HWND, c_int)
SetMagnificationStepTimeArgTypes = ((1, "hwnd"), (1, "MagnificationStepTime"))
GetMagnificationStepTimeFuncType = WINFUNCTYPE(DWORD, HWND, POINTER(c_int))
GetMagnificationStepTimeArgTypes = ((1, "hwnd"), (2, "MagnificationStepTime"))

SetMagnificationSmoothnessFuncType = WINFUNCTYPE(DWORD, HWND, c_int)
SetMagnificationSmoothnessArgTypes = ((1, "hwnd"), (1, "MagnificationSmoothness"))
GetMagnificationSmoothnessFuncType = WINFUNCTYPE(DWORD, HWND, POINTER(c_int))
GetMagnificationSmoothnessArgTypes = ((1, "hwnd"), (2, "MagnificationSmoothness"))

SetColorSchemeFuncType = WINFUNCTYPE(DWORD, HWND, MAGCOLOREFFECT)
SetColorSchemeArgTypes = ((1, "hwnd"), (1, "colorScheme"),)
GetColorSchemeFuncType = WINFUNCTYPE(DWORD, HWND, POINTER(MAGCOLOREFFECT))
GetColorSchemeArgTypes = ((1, "hwnd"), (2, "colorScheme"),)
ShowSystemCursorFuncType = WINFUNCTYPE(DWORD, HWND, BOOL)
ShowSystemCursorArgTypes = ((1, "hwnd"), (1, "showCursor"),)
FollowMouseFuncType = WINFUNCTYPE(DWORD, HWND, BOOL, RECT)
FollowMouseArgTypes = ((1, "hwnd"), (1, "followMouse"), (1, "boundary", RECT(-1,-1,-1,-1)),)
SetBitmapSmoothingFuncType = WINFUNCTYPE(DWORD, HWND, BOOL)
SetBitmapSmoothingArgTypes = ((1, "hwnd"), (1, "useSmoothing"),)
LicensingCallbackFuncType = WINFUNCTYPE(None)

_Initialize = InitializeFuncType(("USA_Initialize", libUSA), InitializeArgTypes)
_Initialize.errcheck = errCheck
_ActivateLicense=ActivateLicenseFuncType(('USA_ActivateLicense', libUSA),((1,'licenseKey'),(1,'success')))
_DeactivateLicense = DeactivateLicenseFuncType(('USA_DeactivateLicense', libUSA),((1,'success'),))
IsLicensed = IsLicensedFuncType(("USA_IsLicensed", libUSA))
Terminate = TerminateFuncType(("USA_Terminate", libUSA))
Terminate.errcheck = errCheck
_CreateMagnifierWindow = CreateMagnifierWindowFuncType(("USA_CreateMagnifierWindow", libUSA), CreateMagnifierWindowArgTypes)
_CreateMagnifierWindow.errcheck = errCheck
_DestroyMagnifierWindow = DestroyMagnifierWindowFuncType(("USA_DestroyMagnifierWindow", libUSA), DestroyMagnifierWindowArgTypes)
_DestroyMagnifierWindow.errcheck =errCheck
SetTrackingRectangle = SetTrackingRectangleFuncType(("USA_SetTrackingRectangle", libUSA), SetTrackingRectangleArgTypes)
SetTrackingRectangle.errcheck = errCheck
GetTrackingRectangle = GetTrackingRectangleFuncType(("USA_GetTrackingRectangle", libUSA), GetTrackingRectangleArgTypes)
GetTrackingRectangle.errcheck = errCheck
SetTrackingOffsets = SetTrackingOffsetsFuncType(("USA_SetTrackingOffsets", libUSA), SetTrackingOffsetsArgTypes)
SetTrackingOffsets.errcheck = errCheck
GetTrackingOffsets = GetTrackingOffsetsFuncType(("USA_GetTrackingOffsets", libUSA), GetTrackingOffsetsArgTypes)
GetTrackingOffsets.errcheck = errCheck
SetMagnificationLevel = SetMagnificationLevelFuncType(("USA_SetMagnificationLevel", libUSA), SetMagnificationLevelArgTypes)
SetMagnificationLevel.errcheck = errCheck
GetMagnificationLevel = GetMagnificationLevelFuncType(("USA_GetMagnificationLevel", libUSA), GetMagnificationLevelArgTypes)
GetMagnificationLevel.errcheck = errCheck
SetMagnificationStepTime = SetMagnificationStepTimeFuncType(("USA_SetMagnificationStepTime_ms", libUSA), SetMagnificationStepTimeArgTypes)
SetMagnificationStepTime.errcheck = errCheck
GetMagnificationStepTime = GetMagnificationStepTimeFuncType(("USA_GetMagnificationStepTime_ms", libUSA), GetMagnificationStepTimeArgTypes)
GetMagnificationStepTime.errcheck = errCheck
SetMagnificationSteps = SetMagnificationStepsFuncType(("USA_SetMagnificationSteps", libUSA), SetMagnificationStepsArgTypes)
SetMagnificationSteps.errcheck = errCheck
GetMagnificationSteps = GetMagnificationStepsFuncType(("USA_GetMagnificationSteps", libUSA), GetMagnificationStepsArgTypes)
GetMagnificationSteps.errcheck = errCheck

GetMagnificationSmoothness = GetMagnificationSmoothnessFuncType(("USA_GetMagnificationSmoothness", libUSA), GetMagnificationSmoothnessArgTypes)
GetMagnificationSmoothness.errcheck = errCheck
SetMagnificationSmoothness = SetMagnificationSmoothnessFuncType(("USA_SetMagnificationSmoothness", libUSA), SetMagnificationSmoothnessArgTypes)
SetMagnificationSmoothness.errcheck = errCheck

SetColorScheme = SetColorSchemeFuncType(("USA_SetColorScheme", libUSA), SetColorSchemeArgTypes)
SetColorScheme.errcheck = errCheck
GetColorScheme = GetColorSchemeFuncType(("USA_GetColorScheme", libUSA), GetColorSchemeArgTypes)
GetColorScheme.errcheck = errCheck
ShowSystemCursor = ShowSystemCursorFuncType(("USA_ShowSystemCursor", libUSA), ShowSystemCursorArgTypes)
ShowSystemCursor.errcheck = errCheck
FollowMouse = FollowMouseFuncType(("USA_FollowMouse", libUSA), FollowMouseArgTypes)
FollowMouse.errcheck = errCheck
SetBitmapSmoothing = SetBitmapSmoothingFuncType(("USA_SetBitmapSmoothing", libUSA), SetBitmapSmoothingArgTypes)
SetBitmapSmoothing.errcheck = errCheck

def Initialize(magnifierType=MAG_TYPE_SCREEN):
	hwnd = _Initialize(magnifierType)
	if magnifierType == MAG_TYPE_SCREEN:
		magWindows[0] = RECT(-1,-1,-1,-1)
	return hwnd

def ActivateLicense(licenseKey):
	success = BOOL()
	message = _ActivateLicense(licenseKey, byref(success))
	return (success, message)

def DeactivateLicense():
	success = BOOL()
	message = _DeactivateLicense(byref(success))
	return (success, message)

def CreateMagnifierWindow(rect):
	hwnd = _CreateMagnifierWindow(rect)
	if hwnd:
		magWindows[hwnd] = rect
	return hwnd

def DestroyMagnifierWindow(hwnd):
	try:
		del magWindows[hwnd]
	except KeyError:
		raise ValueError("hwnd not registered")
	return _DestroyMagnifierWindow(hwnd)

def _setDllFuncPointer(dll,name,cfunc):
	cast(getattr(dll,name),POINTER(c_void_p)).contents.value=cast(cfunc,c_void_p).value if cfunc else None

def SetCallbacks(_onTrial, _onTrialExpired):
	_setDllFuncPointer(libUSA, "_USA_OnTrial", _onTrial)
	_setDllFuncPointer(libUSA, "_USA_OnTrialExpired", _onTrialExpired)

log.info("importing libUSA finished")
