"""
Babbage B.V., 2019-2023

visionEnhancementProviders/_base/__init__.py
implements the base class for USA. The difference between USABase and the USA class in visionEnhancmenetProviders/usa.py
used to be because of a seperate implementation of USA Remote, but these were merged.
"""
# python modules
import wx # type: ignore

# NVDA modules
import vision
from vision.constants import Context
from vision.util import *
import ui
from locationHelper import RectLTRB
import baseObject
import addonHandler
from ctypes.wintypes import RECT
import weakref
import api
from typing import Optional, Type
from logHandler import log


# We need to initialize translation and localization support:
addonHandler.initTranslation()

# not really used (yet)
DEBUG = 1
def debug(string):
	log.debug(string)

class USABase(vision.providerBase.VisionEnhancementProvider, baseObject.ScriptableObject):
	supportedContexts = (
		Context.FOCUS,
		Context.FOREGROUND,
		Context.CARET,
		Context.BROWSEMODE,
		Context.REVIEW,
		Context.NAVIGATOR,
	)
	scriptCategory = _("Ultimate Screen Access") # type: ignore
	_settings = None

	@classmethod
	def getSettingsPanelClass(cls) -> Optional[Type]:
		return None

	@classmethod
	def getSettings(cls):
		return cls._settings

	def __init__(self):
		super().__init__()
		settings = self.getSettings()
		settings.registerProviderInstance(self)
		self.lastCaretObjRef = None
		self.lastReviewMoveContext = None

		self.wordFollower = None
		self.objectReader = None

	def terminate(self):
		self.getSettings().unregisterProviderInstance()
		super().terminate()

	def trackToRectangle(self, rect, context):
		raise NotImplementedError

	def handleUpdate(self, obj):
		if obj is api.getFocusObject():
			self.handleFocusChange(obj)
		elif obj is api.getNavigatorObject():
			self.handleReviewMove(context=Context.NAVIGATOR)

	def handleFocusChange(self, obj):
		context = Context.FOCUS
		self.trackToObject(obj, context=context)
		hasNavigableText = getattr(obj, "_hasNavigableText", False)
		if hasNavigableText:
			# This object most likely has a caret.
			# Intentionally check this after tracking a magnifier to the object itself.
			self.handleCaretMove(obj)

	def handleForegroundChange(self, obj):
		context = Context.FOREGROUND
		self.trackToObject(obj, context=context)

	def handleCaretMove(self, obj):
		self.lastCaretObjRef = weakref.ref(obj)

	def handleReviewMove(self, context):
		if context in (Context.NAVIGATOR, Context.REVIEW):
			self.lastReviewMoveContext = context

	def handleCachedUpdates(self):
		# Prefer caret over review
		handledCaret = False
		if callable(self.lastCaretObjRef):
			obj = self.lastCaretObjRef()
			self.lastCaretObjRef = None
			if obj:
				context = Context.BROWSEMODE if api.isCursorManager(obj) else Context.CARET
				self.trackToObject(obj, context=context)
				handledCaret = True
		if self.lastReviewMoveContext:
			lastReviewMoveContext = self.lastReviewMoveContext
			self.lastReviewMoveContext = None
			if handledCaret:
				return
			self.trackToObject(None, context=lastReviewMoveContext)

	def registerEventExtensionPoints(self, extensionPoints):
		extensionPoints.post_objectUpdate.register(self.handleUpdate)
		extensionPoints.post_focusChange.register(self.handleFocusChange)
		extensionPoints.post_foregroundChange.register(self.handleForegroundChange)
		extensionPoints.post_reviewMove.register(self.handleReviewMove)
		extensionPoints.post_caretMove.register(self.handleCaretMove)
		extensionPoints.post_browseModeMove.register(self.handleCaretMove)
		extensionPoints.post_coreCycle.register(self.handleCachedUpdates)
		self._extensionPoints = extensionPoints

	def unregisterEventExtensionPoints(self):
		if not self._extensionPoints:
			return
		extensionPoints  = self._extensionPoints
		self._extensionPoints = None
		extensionPoints.post_objectUpdate.unregister(self.handleUpdate)
		extensionPoints.post_focusChange.unregister(self.handleFocusChange)
		extensionPoints.post_foregroundChange.unregister(self.handleForegroundChange)
		extensionPoints.post_reviewMove.unregister(self.handleReviewMove)
		extensionPoints.post_caretMove.unregister(self.handleCaretMove)
		extensionPoints.post_browseModeMove.unregister(self.handleCaretMove)
		extensionPoints.post_coreCycle.unregister(self.handleCachedUpdates)

	def _get_enabledContexts(self):
		settings = self.getSettings()
		return tuple(
			context for context in self.supportedContexts
			if getattr(settings, 'trackTo%s' % (context[0].upper() + context[1:]))
		)

	def trackToObject(self, obj, context):
		"""Tracks the magnifier to the given object.
		If object is C{None}, the appropriate object is fetched automatically.
		"""
		log.debug(f"USA: trackToObject {obj} {context}")
		if context not in self.enabledContexts:
			return
		try:
			rect = getContextRect(context, obj)
		except (LookupError, NotImplementedError):
			rect = None
		if not rect:
			return
		self.trackToRectangle(rect, context=context)
		debug(f"USA: trackToObject \n \t obj = {obj} \n \t rect = {rect}")

	def trackToPoint(self, point, context):
		"""Tracks the magnifier to the given point.
		The base implementation creates a rectangle from a point and tracks to that rectangle."""
		log.debug(f"USA: trackToPoint {point} {context}")
		x, y = point
		rect = RectLTRB(x, y, x + 1, y + 1)
		self.trackToRectangle((rect), context=context)
		debug(f"USA: trackToPoint \n \t point = {point}")

	def _setMagnificationLevelHelper(self, increase: bool = True):
		raise NotImplementedError
		level = self.getSettings()._setMagnificationLevelHelper(increase=increase)
		ui.message(_("Magnification level {level}").format(level=level))

	def getContextBoundary(self, context):
		return RECT(*getattr(self.getSettings(), context + "Boundary", (0, 0, 100, 100)))


