"""
Babbage B.V., 2019-2023

visionEhancementProviders/_base/baseSettings.py
Implements some logic on some controls in wx used in the settingspanel of NVDA vision.
The split in abstract USA settings and the regular USA settings in settings.py used to be because of the split in USA
and USA remote, but this was removed to reduce overhead.
"""

from autoSettingsUtils.utils import StringParameterInfo
from vision.providerBase import VisionEnhancementProviderSettings, VisionEnhancementProvider
from vision.constants import Context

from autoSettingsUtils import driverSetting
import addonHandler
from logHandler import log
from .configSpec import configSpec
import weakref
import config
from typing import Optional
from ctypes.wintypes import RECT
import globalPluginHandler
from operator import gt, lt
# We need to initialize translation and localization support:
addonHandler.initTranslation()


_contextOptionLabelsWithAccelerators = {
	Context.FOCUS: _("Follow &focus"),
	Context.FOREGROUND: _("Follow foreground &window changes"),
	Context.NAVIGATOR: _("Follow &navigator object"),
	Context.CARET: _("Follow system &caret"),
	Context.BROWSEMODE: _("Follow &browse mode cursor"),
	Context.REVIEW: _("Follow &review cursor"),
	Context.MOUSE: _("Follow &mouse"),
}

_contextBoundaryLabelsWithAccelerators = {
	Context.FOCUS: _("F&ocus boundary"),
	Context.CARET: _("C&aret boundary"),
	Context.BROWSEMODE: _("Browse mode cursor boundar&y"),
	Context.MOUSE: _("Mouse boun&dary"),
}

_contextBoundaryDefaults = {
	Context.FOCUS: (25, 25, 75, 75),
	Context.CARET: (10, 10, 90, 90),
	Context.BROWSEMODE: (20, 20, 80, 80),
	Context.MOUSE: (10, 10, 90, 90),
}


predefinedStepLevels = (
	0.01, 0.05, 0.1, 0.15, 0.2, 0.25, 0.3, 0.40, 0.5, 0.6, 0.75, 1.0, 1.25, 1.5, 1.75, 2.0
)

predefinedNumberOfMagnificationSteps = (
	1, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 100
)

predefinedMagnificationStepTime_ms = (
	1, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 70, 75, 80, 85, 90, 95, 100
)

predefinedMagnificationLevels = (
	1.0, 1.25, 1.5, 1.75,
	2.0, 2.5, 3.0,
	4.0, 5.0, 6.5, 8.0,
	10.0, 12.5, 15.0, 17.5,
	20.0, 25.0, 30.0, 37.5,
	45.0, 52.5, 60.0, 70.0, 80.0
)

predefinedZoomFrames = (
	1,2,3,4, 5, 6, 7, 8, 9, 10
)

predefinedMagnificationSmoothness = (
	1,10,20,30,40,50,60,70,80,90,100
)

class TupleDriverSetting(driverSetting.DriverSetting):

	def __init__(
			self,
			id,
			displayNameWithAccelerator,
			defaultVal,
			min: int = 4,
			max: int = 4,
			displayName: Optional[str] = None,
			useConfig: bool = True
	):
		super().__init__(
			id,
			displayNameWithAccelerator,
			availableInSettingsRing=False,
			defaultVal=defaultVal,
			displayName=displayName,
			useConfig=useConfig
		)
		self.min = min
		self.max = max

	def _get_configSpec(self):
		return f"int_list(default=list{self.defaultVal}, min={self.min}, max={self.max})"

class FloatDriverSetting(driverSetting.DriverSetting):

	def _get_configSpec(self):
		return f"float(default={self.defaultVal})"


class StringWithValueParameterInfo(StringParameterInfo):
	def __init__(self, id, name, value):
		super().__init__(id, name)
		self.value = value

class AbstractUSASettings(VisionEnhancementProviderSettings):

	@classmethod
	def _get_conf(cls):
		return config.conf[cls._getConfigSection()][cls.getId()]

	def __init__(self):
		self._providerInstRef = lambda: None
		self.cachedMagnificationLevel = None
		super().__init__()

	def registerProviderInstance(self, inst: VisionEnhancementProvider):
		self._providerInstRef = weakref.ref(inst)
		# ensure any previously saved settings are loaded from config file:
		self._initSpecificSettings(self, self._getAvailableRuntimeSettings())

	def unregisterProviderInstance(self):
		if self._providerInstRef():
			self._saveSpecificSettings(self, self._getAvailableRuntimeSettings())
		self._providerInstRef = lambda: None


	def _get_supportedSettings(self):
		if not self._providerInstRef():
			return []
		return self._getAvailableRuntimeSettings()

	def _getAvailableRuntimeSettings(self):
		settings = [
			FloatDriverSetting(
				'magnificationLevel',
				_("Magnification &level"),
				defaultVal=1.0,

			),
			FloatDriverSetting(
				'stepLevel',
				_("step &level"),
				defaultVal=0.1,

			),			
		]
		return settings

	_abstract_magnificationLevel = True

	def _get_magnificationLevel(self):
		...

	def _get_setMagnificationByLevel(self):
		...

	def _get_MagnificationSteps(self):
		...

	def _get_MagnificationStepTime_ms(self):
		...

	def _setMagnificationByStepHelper(self, increase=True):
		# hardcoded the minimum and maximum level
		maxLevel = 100
		minLevel = 1
		if self.cachedMagnificationLevel:
			level = self.cachedMagnificationLevel
			self.cachedMagnificationLevel = None
		else:
			curLevel = self.magnificationLevel
			level = curLevel * (1 + self.stepLevel) if increase else curLevel / (1 + self.stepLevel)
			level = max(min(maxLevel, level), minLevel)
		self.magnificationLevel = level
		return level

	def _setMagnificationByLevelHelper(self, increase=True):
		if self.cachedMagnificationLevel:
			level = self.cachedMagnificationLevel
			self.cachedMagnificationLevel = None
		else:
			curLevel = self.magnificationLevel
			levels = (iter if increase else reversed)(predefinedMagnificationLevels)
			for level in levels:
				if (gt if increase else lt)(level, curLevel):
					break
			else:
				level = curLevel
		self.magnificationLevel = level
		return level

	def _setMagnificationLevelHelper(self, increase=True):
		if self.setMagnificationByLevel:
			return self._setMagnificationByLevelHelper(increase)
		else:
			return self._setMagnificationByStepHelper(increase)

	
	def _setMagnificationSmoothnessHelper(self, increase=True):
		curLevel = self.MagnificationSmoothness
		levels = (iter if increase else reversed)(predefinedMagnificationSmoothness)
		for level in levels:
			if (gt if increase else lt)(level, curLevel):
				self.MagnificationSmoothness = level
				break
		else:
			self.MagnificationSmoothness = curLevel
		return self.MagnificationSmoothness


	def _toggleMagnifierHelper(self):
		if self.cachedMagnificationLevel:
			level = self.magnificationLevel = self.cachedMagnificationLevel
			self.cachedMagnificationLevel = None
			return True
		else:
			self.cachedMagnificationLevel = self.magnificationLevel
			self.magnificationLevel = 1.0
			return False

	@classmethod
	def _get_availableMagnificationlevels(self):
		return {
			val: StringParameterInfo(val, str(val))
			for val in predefinedMagnificationLevels
		}

	@classmethod
	def _get_availableSteplevels(self):
		return {
			val: StringParameterInfo(val, str(val))
			for val in predefinedStepLevels
		}

	@classmethod
	def _get_availableMagnificationstepss(self):
		return {
			val: StringParameterInfo(val, str(val))
			for val in predefinedNumberOfMagnificationSteps
		}

	@classmethod
	def _get_availableMagnificationsteptime_mss(self):
		return {
			val: StringParameterInfo(val, str(val))
			for val in predefinedMagnificationStepTime_ms
		}

	@classmethod
	def _get_availableMagnificationsmoothnesss(self):
		return {
			val: StringParameterInfo(val, str(val))
			for val in predefinedMagnificationSmoothness
		}


