"""
Babbage B.V., 2019-2023

visionEnhancementProviders/_base/settings.py
contains the main code for the settings of the USA vision enhancement provider.
"""

from sys import exc_info
import wx # type: ignore
from .baseSettings import *
from .baseSettings import _contextOptionLabelsWithAccelerators, _contextBoundaryLabelsWithAccelerators, _contextBoundaryDefaults
from .libUSA import libUSA

# We need to initialize translation and localization support:
addonHandler.initTranslation()



def getScreenEdgesString(i):
	# Translator: xgettext:no-python-format
	return _("within {i}% from screen edges").format(i=i)


class USASettings(AbstractUSASettings):

	@classmethod
	def getId(cls) -> str:
		return "usa"

	@classmethod
	def getDisplayName(cls) -> str:
		return "Ultimate Screen Access"

	def initSettings(self):
		super().initSettings()

		# transfer the strings in customColorScheme to proper floats, since it is saved as strings
		if config.conf[self._getConfigSection()][self.getId()].get('customColorScheme'):
			for key in ['r','g','b','a','t']:
				jj = [eval(ii) for ii in config.conf[self._getConfigSection()][self.getId()]['customColorScheme'][key]   ]
				config.conf[self._getConfigSection()][self.getId()]['customColorScheme'][key] = jj

		# fgColorschemes is not properly initialized the first time, so it is here
		if not config.conf[self._getConfigSection()][self.getId()].get('fgColorschemes'):
			config.conf[self._getConfigSection()][self.getId()]['fgColorschemes'] = {}
		else:
			# if we already have a aggerated section than we need to set the 'None'strings to actual None
			for key, item in list(config.conf[self._getConfigSection()][self.getId()]['fgColorschemes'].items()):

				if item == 'None':
					config.conf[self._getConfigSection()][self.getId()]['fgColorschemes'][key] = None
				# config.conf[self._getConfigSection()][self.getId()]['fgColorschemes'][key] = ii

		# update form the configSpec file if needed
		config.conf[self._getConfigSection()][self.getId()].spec.update(configSpec)


	def _getAvailableRuntimeSettings(self):

		settings = super()._getAvailableRuntimeSettings()
		for context, label in _contextOptionLabelsWithAccelerators.items():
			context = context.value
			defaultValue = not context == Context.NAVIGATOR.value
			settings.append(
				driverSetting.BooleanDriverSetting(
					f'trackTo{context[0].upper()}{context[1:]}',
					label,
					defaultVal=defaultValue
				)
			)
			if context in _contextBoundaryLabelsWithAccelerators:
				settings.append(
					TupleDriverSetting(
						f'{context}Boundary',
						_contextBoundaryLabelsWithAccelerators[context],
						defaultVal=_contextBoundaryDefaults[context]
					)
				)
		settings.extend([
			driverSetting.BooleanDriverSetting(
				"bitmapSmoothing",
				_("Enable bitmap &smoothing to increase magnification quality"),
				defaultVal=True
			),
			driverSetting.BooleanDriverSetting(
				"showSystemCursor",
				_("Show ma&gnified system cursor"),
				defaultVal=True
			),
			driverSetting.DriverSetting(
				'colorScheme',
				_("Color sc&heme"),
				defaultVal="default"
			),
			driverSetting.BooleanDriverSetting(
				"followWord",
				_("Enable following of words in continuous reading mode."),
				defaultVal=True
			),
			# settings bool used for rectHighlighter
			driverSetting.BooleanDriverSetting(
				"followWordRect",
				_("Enable highlighting of words in continuous reading mode."),
				defaultVal=True
			),
			driverSetting.BooleanDriverSetting(
				"followWordGlobal",
				_("Enable following of words everywhere."),
				defaultVal=False
			),
			# settings bool used for rectHighlighter
			driverSetting.BooleanDriverSetting(
				"followWordJumpCaret",
				_("Let the caret jump with the read words in continuous reading mode."),
				defaultVal=False
			),
			driverSetting.DriverSetting(
				'caretColor',
				_("Wordfollower rectangle color"),
				defaultVal="Blue"
			),
			# settings bool used for rectHighlighter
			driverSetting.BooleanDriverSetting(
				"setMagnificationByLevel",
				_("Let USA set the magnification level by fixed levels"),
				defaultVal=True
			),
			driverSetting.BooleanDriverSetting(
				"autoStartUSAWhenOff",
				_("Let USA start automatically when it is off and you start NVDA"),
				defaultVal=False
			),
			FloatDriverSetting(
				"MagnificationSmoothness",
				_("How smoothly the screen moves when zooming or moving the screen"),
				defaultVal=70
			)
		])
		return settings
		# return sorted list of settings based on what it does
		#return sorted(settings, key=lambda x: x._get_configSpec() )


	_availableBoundaries = {
		val: StringParameterInfo(list(val), name)
		for val, name in {
			(0, 0, 100, 100): _("unrestricted"),
			(5, 5, 95, 95): getScreenEdgesString(5),
			(10, 10, 90, 90): getScreenEdgesString(10),
			(15, 15, 85, 85): getScreenEdgesString(15),
			(20, 20, 80, 80): getScreenEdgesString(20),
			(25, 25, 75, 75): getScreenEdgesString(25),
			(30, 30, 70, 70): getScreenEdgesString(30),
			(35, 35, 65, 65): getScreenEdgesString(35),
			(40, 40, 60, 60): getScreenEdgesString(40),
			(45, 45, 55, 55): getScreenEdgesString(45),
			(50, 50, 50, 50): _("centered on screen"),
			(-1, -1, -1, -1): _("relative"),
		}.items()
	}

	availableFocusboundarys = availableCaretboundarys = availableBrowsemodeboundarys = availableMouseboundarys = _availableBoundaries

	_followWord: bool = True
	def _get_followWord(self):
		return self._followWord

	def _set_followWord(self, bool):
		self._followWord = bool

	_followWordGlobal: bool = False
	def _get_followWordGlobal(self):
		return self._followWordGlobal

	def _set_followWordGlobal(self, bool):
		self._followWordGlobal = bool

	_followWordRect: bool = True
	def _get_followWordRect(self):
		return self._followWordRect
	
	def _set_followWordRect(self, bool):
		self._followWordRect = bool

	_setMagnificationByLevel: bool = True
	def _get_setMagnificationByLevel(self):
		return self._setMagnificationByLevel

	def _set_setMagnificationByLevel(self, bool):
		self._setMagnificationByLevel = bool

	
	def _get_autoStartUSAWhenOff(self):
		return self.conf.get("autoStartUSAWhenOff") == "True" or self.conf.get("autoStartUSAWhenOff") == True
	
	def _set_autoStartUSAWhenOff(self, bool):
		self.conf["autoStartUSAWhenOff"] = bool
		#self._autoStartUSAWhenOff = bool

	_followWordJumpCaret: bool = True
	def _get_followWordJumpCaret(self):
		return self._followWordJumpCaret

	def _set_followWordJumpCaret(self, bool):
		self._followWordJumpCaret = bool
	def _get_isRemote(self):
		if self.conf.get("isRemote") == "1" or self.conf.get("isRemote") == 1:
			return True
		else:
			return False

	def _set_isRemote(self, bool):
		if bool:
			self.conf["isRemote"] = "1"
		else:
			self.conf["isRemote"] = "0"

	def _get_objectReaderMode(self):
		if self.conf.get("objectReaderMode") == "1" or self.conf.get("objectReaderMode") == 1:
			return True
		else:
			return False

	def _set_objectReaderMode(self, bool):
		if bool:
			self.conf["objectReaderMode"] = "1"
		else:
			self.conf["objectReaderMode"] = "0"

	def _get_availableCaretcolors(self):

		return {
			val: StringParameterInfo(val, str(val))
			for val in ["Blue", "Pink", "Yellow"]
		}


	def _get_remoteHandler(self):
		plugin = next((p for p in globalPluginHandler.runningPlugins if p.__module__=='globalPlugins.usa'), None)
		remoteHandler = plugin.remoteHandler if plugin else None
		return remoteHandler


	_remoteMagnificationLevel = 1.0
	def _get_magnificationLevel(self):
		if self.isRemote:
			return self._remoteMagnificationLevel
		else:
			return libUSA.GetMagnificationLevel(list(libUSA.magWindows.keys())[0])

	def _set_magnificationLevel(self, level):
		if self.isRemote:
			pass
		else:
			for hwnd in libUSA.magWindows:
				try:
					libUSA.SetMagnificationLevel(hwnd, level)
				except WindowsError:
					log.error("Couldn't set magnification level", exc_info=True)


	_remoteMagnificationSteps: int = 1
	def _get_MagnificationSteps(self) -> int:
		# Number of substeps that the magnifier should take between the current zoomframe and desired zoomframe
		if self.isRemote:
			return self._remoteMagnificationSteps
		else:
			return libUSA.GetMagnificationSteps(list(libUSA.magWindows.keys())[0])

	def _set_MagnificationSteps(self, MagnificationSteps: int):
		# Number of substeps that the magnifier should take between the current zoomframe and desired zoomframe
		if self.isRemote:
			pass
		else:
			for hwnd in libUSA.magWindows:
				try:
					libUSA.SetMagnificationSteps(hwnd, int(MagnificationSteps))
				except WindowsError:
					log.error("Couldn't set number of magnification steps", exc_info=True)

	_remoteMagnificationStepTime_ms: int = 1
	def _get_MagnificationStepTime_ms(self) -> int:
		# Amount of ms between each substep that is taken by the magnifier between the current zoomframe and desired zoomframe
		if self.isRemote:
			return self._remoteMagnificationStepTime_ms
		else:
			return libUSA.GetMagnificationStepTime(list(libUSA.magWindows.keys())[0])

	def _set_MagnificationStepTime_ms(self, MagnificationStepTime_ms: int):
		# Amount of ms between each substep that is taken by the magnifier between the current zoomframe and desired zoomframe
		if self.isRemote:
			pass
		else:
			for hwnd in libUSA.magWindows:
				try:
					libUSA.SetMagnificationStepTime(hwnd, int(MagnificationStepTime_ms))
				except WindowsError:
					log.error("Couldn't set time of magnification step", exc_info=True)

	_remoteMagnificationSmoothness: int = 1
	def _get_MagnificationSmoothness(self):
		if self.isRemote:
			return self._remoteMagnificationSmoothness
		else:
			return libUSA.GetMagnificationSmoothness(list(libUSA.magWindows.keys())[0])

	def _set_MagnificationSmoothness(self, MagnificationSmoothness: int):
		if self.isRemote:
			pass
		else:
			for hwnd in libUSA.magWindows:
				try:
					libUSA.SetMagnificationSmoothness(hwnd, int(MagnificationSmoothness))
				except WindowsError:
					log.error("Couldn't set magnification smoothness", exc_info=True)

	_bitmapSmoothing: bool = True
	def _get_bitmapSmoothing(self):
		return self._bitmapSmoothing

	def _set_bitmapSmoothing(self, state):
		if self.isRemote:
			pass
		else:
			for hwnd in libUSA.magWindows:
				try:
					libUSA.SetBitmapSmoothing(hwnd, state)
				except WindowsError:
					log.error("Bitmap smoothing not available for window %s" % hwnd, exc_info=True)
					return
			self._bitmapSmoothing = state

	def _get_colorScheme(self):
		if not self.conf.get('colorScheme'):
			self.conf['colorScheme'] = 'default'
		return self.conf['colorScheme']

	def _set_colorScheme(self, transformation):
		if self.isRemote:
			pass
		else:
			try:
				info = self.availableColorschemes[transformation]
			except KeyError:
				log.error("Transformation not in list! Continuing with first item in list")
				# this is a fail safe which should not have to be used. If the error above shows up, than you should definetly fix something
				firstKey = list(self.availableColorschemes.keys())[0]
				info = self.availableColorschemes.get(firstKey)
			for hwnd in libUSA.magWindows:
				try:
					libUSA.SetColorScheme(hwnd, info.value)
				except WindowsError:
					log.error("Couldn't set color scheme", exc_info=True)
					return
			self.conf['colorScheme'] = transformation

	_showSystemCursor: bool = True
	def _get_showSystemCursor(self):
		return self._showSystemCursor

	def _set_showSystemCursor(self, state):
		if self.isRemote:
			pass
		else:
			for hwnd in libUSA.magWindows:
				try:
					libUSA.ShowSystemCursor(hwnd, state)
				except WindowsError:
					log.error("Couldn't change show system cursor for window %s" % hwnd, exc_info=True)
					return
			self._showSystemCursor = state

	_trackToMouse: bool = True
	def _get_trackToMouse(self):
		return self._trackToMouse

	def _set_trackToMouse(self, state):
		if self.isRemote:
			pass
		else:
			self._trackToMouse = state
			# the mouse boundary is set after this settings, so to give USA a change to give the right value to
			# the usa dll, we call the mousetoggle afterwardsn
			wx.CallLater(1000, self.toggleMouseFollow)


	def toggleMouseFollow(self):
		for hwnd in libUSA.magWindows:
			try:
				libUSA.FollowMouse(hwnd, self._trackToMouse, RECT(*self.mouseBoundary))
			except WindowsError:
				log.error("Follow mouse not available for window %s" % hwnd, exc_info=True)
				return

	@classmethod
	def _get_allColorschemes(cls):
		# if customColorScheme is not defined yet, define it as follows:
		# this was previously in configSpec but moved to here since the latter did not work properly.
		if not cls.conf.get('customColorScheme'):
			cls.conf['customColorScheme'] = {
				'r': [1.0, 0.0, 0.0, 0.0, 0.0],
				'g': [0.0, 1.0, 0.0, 0.0, 0.0],
				'b': [0.0, 0.0, 1.0, 0.0, 0.0],
				'a': [0.0, 0.0, 0.0, 1.0, 0.0],
				't': [0.0, 0.0, 0.0, 0.0, 1.0],
			}
		transformations = [
			StringWithValueParameterInfo("default", _("True colors"), libUSA.MAGCOLOREFFECT((
				(1.0, 0.0, 0.0, 0.0, 0.0),
				(0.0, 1.0, 0.0, 0.0, 0.0),
				(0.0, 0.0, 1.0, 0.0, 0.0),
				(0.0, 0.0, 0.0, 1.0, 0.0),
				(0.0, 0.0, 0.0, 0.0, 1.0)
			))),
			StringWithValueParameterInfo("inverted", _("Inverted colors"), libUSA.MAGCOLOREFFECT((
				(-1.0, 0.0, 0.0, 0.0, 0.0),
				(0.0, -1.0, 0.0, 0.0, 0.0),
				(0.0, 0.0, -1.0, 0.0, 0.0),
				(0.0, 0.0, 0.0, 1.0, 0.0),
				(1.0, 1.0, 1.0, 0.0, 1.0)
			))),
			StringWithValueParameterInfo("gray", _("Gray scale"), libUSA.MAGCOLOREFFECT((
				(0.3, 0.3, 0.3, 0.0, 0.0),
				(0.6, 0.6, 0.6, 0.0, 0.0),
				(0.1, 0.1, 0.1, 0.0, 0.0),
				(0.0, 0.0, 0.0, 1.0, 0.0),
				(0.0, 0.0, 0.0, 0.0, 1.0)
			))),
			StringWithValueParameterInfo("highContrast", _("High contrast"), libUSA.MAGCOLOREFFECT((
				(2.0, 0.0, 0.0, 0.0, 0.0),
				(0.0, 2.0, 0.0, 0.0, 0.0),
				(0.0, 0.0, 2.0, 0.0, 0.0),
				(0.0, 0.0, 0.0, 1.0, 0.0),
				(-0.7, -0.7, -0.7, 0.0, 1.0)
			))),
			StringWithValueParameterInfo("invertedHighContrast", _("Inverted high contrast"),
				libUSA.MAGCOLOREFFECT((
				(-2.0, 0.0, 0.0, 0.0, 0.0),
				(0.0, -2.0, 0.0, 0.0, 0.0),
				(0.0, 0.0, -2.0, 0.0, 0.0),
				(0.0, 0.0, 0.0, 1.0, 0.0),
				(1.7, 1.7, 1.7, 0.0, 1.0)
			))),
			StringWithValueParameterInfo("blueLight", _("Blue light"),
				libUSA.MAGCOLOREFFECT((
					(0.8, 0.0, 0.0, 0.0, 0.0),
					(0.0, 0.5, 0.0, 0.0, 0.0),
					(0.0, 0.0, 0.0, 0.0, 0.0),
					(0.0, 0.0, 0.0, 1.0, 0.0),
					(0.0, 0.0, 0.0, 0.0, 1.0)
			))),
			StringWithValueParameterInfo("custom", _("Custom"), libUSA.MAGCOLOREFFECT((
				tuple(cls.conf['customColorScheme']['r']),
				tuple(cls.conf['customColorScheme']['g']),
				tuple(cls.conf['customColorScheme']['b']),
				tuple(cls.conf['customColorScheme']['a']),
				tuple(cls.conf['customColorScheme']['t'])
			))),
		]
		return {transformation.id: transformation for transformation in transformations}

	def _get_availableColorschemes(self):
		# this function looks at teh settings and makes a list of all possible transformations. If none are selected,
		# all are returned
		allTransformations = self._get_allColorschemes()
		selectedTransformations = {}
		if self.conf.get("selectedDefault") == 'True': selectedTransformations["default"] = allTransformations[
			"default"]
		if self.conf.get("selectedInverted") == 'True': selectedTransformations["inverted"] = allTransformations[
			"inverted"]
		if self.conf.get("selectedGray") == 'True': selectedTransformations["gray"] = allTransformations["gray"]
		if self.conf.get("selectedHighContrast") == 'True': selectedTransformations["highContrast"] = \
		allTransformations["highContrast"]
		if self.conf.get("selectedBlueLight") == 'True': selectedTransformations["blueLight"] = allTransformations["blueLight"]
		if self.conf.get("selectedInvertedHighContrast") == 'True': selectedTransformations[
			"invertedHighContrast"] = allTransformations["invertedHighContrast"]
		if self.conf.get("selectedCustom") == 'True': selectedTransformations["custom"] = allTransformations[
			"custom"]

		if not selectedTransformations:
			selectedTransformations["default"] = allTransformations["default"]
			selectedTransformations["inverted"] = allTransformations["inverted"]
			selectedTransformations["gray"] = allTransformations["gray"]
			selectedTransformations["highContrast"] = allTransformations["highContrast"]
			selectedTransformations["invertedHighContrast"] = allTransformations["invertedHighContrast"]
			selectedTransformations["blueLight"] = allTransformations["blueLight"]
			selectedTransformations["custom"] = allTransformations["custom"]
			return allTransformations

		return selectedTransformations



