
"""
2021, Babbage B.V.
Extension to the USA add-on to let USA follow spoken words with a SayAll that starts where the mouse is.
known as "klik&lees"
"""

import wx # type: ignore
from mouseHandler import *
import speech.sayAll
import globalCommands
import virtualBuffers
import api

def isObjectInForegroundObject(obj):
	"""
	iteratively goes through all parents of the object to see whether the object is contained by the current foreground object in NVDA
	"""
	fg = api.getForegroundObject()
	foundFGinParents = fg == obj
	while not foundFGinParents:
		obj = obj.parent
		foundFGinParents = fg == obj
		if not obj:
			# no more object to take the parent from
			return False
	return foundFGinParents

class objectReader:
	"""
	The object reader Mode turns off speech by default. With this you can let NVDA read whatever is under the mouse with a
	simple command.
	"""
	def __init__(self, USAinstance):
		# keep reference to USA
		self.USA = USAinstance

		self.enabled = False
		if self.USA.getSettings().objectReaderMode:
			self.enable()

		self.isSelecting = False
		self.oldFollowMouseValue = False
		self.oldenableMouseTracking = False
		self.vbuf = None

	def terminate(self):
		if self.enabled:
			self.disable()
	def enable(self):
		log.info("USA: enable object Reader Mode")
		speech.speech.setSpeechMode(speech.speech.SpeechMode.off)
		self.enabled = True

	def disable(self):
		log.info("USA: disable object Reader Mode")
		self.enabled = False

	def beepOnError(self):
		# function is not used yet but can be used to indicate non-intuitive errors of the object reader by beeping or messaging
		pass

	def handleForegroundChange(self, obj):
		pass

	def handleReviewMove(self, context):
		pass

	def handleFocusChange(self, obj):
		pass

	def callAfterCommandsForMouseRead(self):
		log.debugWarning("USA: objectReader; callAfterCommandsForMouseRead called")
		speech.speech.setSpeechMode(speech.speech.SpeechMode.talk)

		wx.CallLater(100, speech.sayAll.SayAllHandler.readText, speech.sayAll.CURSOR.CARET)
		wx.CallLater(300, self.disableVbuf)

	def reloadVbuf(self):
		# we have a virtual buffer where we need to set the caret on the right position.
		log.debugWarning("USA: objectReader; reloadVbuf called")
		desktop = api.getDesktopObject()
		x, y = winUser.getCursorPos()
		obj = desktop.objectFromPoint(x, y)
		self.vbuf = self.setCaretOnRelevantTextInfo(obj)
		log.debugWarning(f"USA: objectReader; self.vbuf: {self.vbuf}")

	def firstVbufLoad(self):
		# we have a virtual buffer where we need to set the caret on the right position, but we have to initialize it first.
		log.debugWarning("USA: objectReader; firstVbufLoad called")
		globalCommands.commands.script_toggleVirtualBufferPassThrough("")
		wx.CallLater(200, self.reloadVbuf)
		wx.CallLater(400, self.callAfterCommandsForMouseRead)

	def readAllUnderMouse(self):
		# initial call when the command to say all under mouse is pressed.

		#get current object under the mouse. This can be different than the mouseObject in NVDA itself.
		x, y = winUser.getCursorPos()
		desktop = api.getDesktopObject()
		obj = desktop.objectFromPoint(x, y)

		if not obj:
			log.error("USA: objectReader; could not get object under the mouse! Returning early")
			self.beepOnError()
			return

		if not isObjectInForegroundObject(obj):
			log.error("USA: objectReader; object under mouse is not part of current active foreground object! Returning early")
			self.beepOnError()
			return

		log.debugWarning(f"USA: objectReader; readAllUnderMouse called with object {obj}")
		vbufClass = self.getTIClass(obj)
		if vbufClass and issubclass(vbufClass, (virtualBuffers.VirtualBuffer,) ):
			# if we have a tree interceptor that is a virtual buffer, then we first need to place the caret in the right
			# position.
			vbuf = obj.treeInterceptor

			if not vbuf:
				log.debugWarning("USA: objectReader; vbuf needs to be made first")
				# the vbuf needs to be made first. Everything else needs to be called after. Thus we return early here
				self.firstVbufLoad()
				return
			else:
				# activate the already initialized vbuf
				log.debugWarning("USA: objectReader; toggle vbuf")
				if vbuf.passThrough:
					vbuf.passThrough = False
					vbuf.disableAutoPassThrough = False
			self.reloadVbuf()

		self.callAfterCommandsForMouseRead()

	def disableVbuf(self):
		log.debugWarning("USA: objectReader; disableVbuf called")
		if self.vbuf:
			self.vbuf.passThrough = False
			self.vbuf.disableAutoPassThrough = False
		#self.vbuf = None

	def setCaretOnRelevantTextInfo(self, obj):

		if not obj.appModule.helperLocalBindingHandle:
			# at the very start of NVDA it is possible that the nvdaHelperRemote.dll is not injected yet. This causes that
			# NVDA cannot put a caret down anyway.
			log.error(f"called setCaretOnRelevantTextInfo but {obj.appModule} does not have a nvdaHelperRemote yet")
			self.beepOnError()
			return

		vbuf = obj.treeInterceptor
		if not vbuf:
			# the vbuf should have been initialized when this function was called. The code can come here when we recursively
			# go through the parents to try and set te caret on the correct textInfo object. When we go above the rootNVDAObject
			# of the treeInterceptor, we will not have a treeInterceptor and our attempt to put the caret somewhere reasonable
			# has failed.
			log.error(f"called setCaretOnRelevantTextInfo but {obj} does not have a treeInterceptor instance.")
			self.beepOnError()
			return

		log.debugWarning(f"USA: ObjectReader; setCaretOnRelevantTextInfo obj {obj} vbuf {vbuf}")
		try:
			# first try to make a textInfo. the embedded try-except structure is just to differentiate between errors in
			# makeTextInfo and updateCaret
			try:
				info = vbuf.makeTextInfo(obj)
				log.debugWarning(f"USA: ObjectReader;  made info object {info}")
			except Exception as e:
				log.error(f"USA: ObjectReader;  vbuf.makeTextInfo(obj) error with  obj {obj} and vbuf {vbuf}. Original error: \n{e}")
				raise e

			# if we have a textInfo object, try and put the caret on the textInfo object that we got. If this fails we try
			# the text object of the parent.
			info.updateCaret()
			log.debugWarning(f"USA: ObjectReader;  set caret to {info} \n caret object: {api.getCaretObject()}")
			return vbuf
		except Exception as e:
			# keep trying to put the caret on the textinfo object of the parent
			log.debugWarning(f"USA: ObjectReader;  failed setCaretOnRelevantTextInfo because {e}")
			parent = obj.parent
			if parent:
				return self.setCaretOnRelevantTextInfo(parent)

		# If we are here it means that we went through the whole tree up to the desktopObject. thus we did not find a vbuf
		# at all while we did except one to.
		log.error(f"USA: ObjectReader; failed setCaretOnRelevantTextInfo, no vbuf found at all")
		self.beepOnError()

	def disableSpeechAndVbuf(self):
		speech.speech.setSpeechMode(speech.speech.SpeechMode.off)
		log.debugWarning(f"USA: ObjectReader; disableSpeechAndVbuf")
		self.USA.wordFollower.rectHighlighter.updateCustomRect(rect=None)
		if self.vbuf:
			log.info("disabling vbuf passThrough")
			self.vbuf.passThrough = True
			self.vbuf.disableAutoPassThrough = True

	def pullTheHandBreak(self):
		log.debugWarning(f"USA: ObjectReader; pullTheHandBreak")
		speech.speech.cancelSpeech()
		speech.speech.setSpeechMode(speech.speech.SpeechMode.off)

		self.USA.wordFollower.rectHighlighter.updateCustomRect(rect=None)
		log.info(f"checkIfSayAllIsRunning stopped {self.vbuf}")
		if self.vbuf:
			log.info("disabling vbuf passThrough")
			self.vbuf.passThrough = True
			self.vbuf.disableAutoPassThrough = True


	def getTIClass(self, obj):
		# gets the first Tree Interceptor of the object, if there is any
		vbufClass = None
		while not vbufClass:

			if obj: # and not obj.role == controlTypes.Role.WINDOW:
				try:
					vbufClass = obj.treeInterceptorClass
				except:
					log.debugWarning(f"USA: ObjectReader; getTIClass NO VBUFCLASS FOUND, check parent")
					vbufClass = None
				obj = obj.parent
			else:
				log.debugWarning(f"USA: ObjectReader; getTIClass NO VBUFCLASS FOUND")
				return None

		return vbufClass

	def toggleVBuf(self, obj):

		vBufClass = self.getTIClass(obj)
		vbuf = None
		if vBufClass:

			if issubclass(vBufClass, (virtualBuffers.VirtualBuffer,) ):
				log.debugWarning("USA: objectReader; activate vbuf")
				vbuf = self.activateVBuf(obj)
			else:
				log.debugWarning("USA: objectReader; deactivate vbuf")
				vbuf = self.deactivateVBuf(obj)

		return vbuf


