"""
2021, Babbage B.V.
Extension to the USA add-on to let USA follow spoken words during SayAll
"""
import speech

import textInfos
import ui
from logHandler import log
from speech import sayAll
from vision.constants import Context
from . import newSpeech, rectHighlighter, textDifference, textProcessor

try:
    from speech.extensions import speechCanceled  # type: ignore
except ImportError:
    speechCanceled = None


class wordFollower:
    """
    General entrance to the functionality of following words while reading text
    """

    def log(self, string : str):
        log.debugWarning(f"Wordfollower USA: {string}")

    def __init__(self, USAinstance):
        self.USA = USAinstance
        self.enabled = False
        self.rectHighlighter = None
        self.oldGetTextInfoSpeech = speech.getTextInfoSpeech

        if self.USA.getSettings().followWord or self.USA.getSettings().followWordGlobal:
            self.enable_wordFollower(notify=False)

    def terminate(self):
        self.log("terminating wordFollower")
        # disable itself if enabled
        if self.enabled:
            self.disable_wordFollower(notify=False)
        if self.rectHighlighter:
            self.rectHighlighter.terminate()
        self.log("wordFollower terminated")

    def enable_wordFollower(self, notify = True):
        self.log("USA: enable_wordFollower")

        sayAll.SayAllHandler._getTextInfoSpeech = newSpeech.getTextInfoSpeech
        if self.USA.getSettings().followWordGlobal:
            speech.speech.getTextInfoSpeech = newSpeech.getTextInfoSpeech
        newSpeech.wordRead.register(self.onCallBack)
        if speechCanceled: speechCanceled.register(self.speechCanceled)
        if self.USA.getSettings().followWordRect:
            self.enable_wordFollowerRect(notify=False)
        self.enabled = True
        if notify:
            ui.message(_('Word follower for USA enabled'))

    def disable_wordFollower(self, notify = True):
        log.info("USA: disable_wordFollower")
        self.disable_wordFollowerRect(notify=False)
        if speechCanceled: speechCanceled.unregister(self.speechCanceled)
        # reset overwritten functions
        speech.speech.getTextInfoSpeech = self.oldGetTextInfoSpeech
        sayAll.SayAllHandler._getTextInfoSpeech = speech.getTextInfoSpeech
        self.enabled = False
        if notify:
            ui.message(_('Word follower for USA disabled'))

    def handleForegroundChange(self, obj):
        if self.rectHighlighter:
            self.rectHighlighter.updateCustomRect(None)

    def handleReviewMove(self, context):
        pass

    def speechCanceled(self, *args, **kwargs):
        if self.rectHighlighter:
            self.rectHighlighter.updateCustomRect(None)

    def handleFocusChange(self, obj):
        if self.rectHighlighter:
            self.rectHighlighter.updateCustomRect(None)

    def enable_wordFollowerRect(self, notify = True):
        log.info("USA: enable_wordFollowerRect")
        if self.USA.getSettings().followWord:
            self.rectHighlighter = rectHighlighter.rectHighlighter(self.USA)
            if notify:
                ui.message(_('Word follower rectangle for USA enabled'))

    def disable_wordFollowerRect(self, notify = True):
        log.info("USA: disable_wordFollowerRect")
        if self.rectHighlighter:
            self.rectHighlighter.terminate()
            self.rectHighlighter = None
            if notify:
                ui.message(_('Word follower rectangle for USA disabled'))

    def correctForMissalignment(self, reader, originalText, bookmark):
        """
        In vbufs there is a possibility that the text changes between the making of the callback and the calling of the
        callback. If that occurs the offsets previously calculated do not work anymore. A function is called that looks
        at the difference between the original text and the current text and tries to deduct what position it should
        be in.
        """
        self.log("correction for potential misallignment")
        # in virtual Buffers the text can change and thus the location of original word that should be spoken.
        # we check what is different between the original and current text and correct the bookmark for it.

        # get the current text
        reader.move(textInfos.UNIT_STORY, direction=-1, endPoint='start')
        reader.move(textInfos.UNIT_STORY, direction=1, endPoint='end')

        # get list with all matches of the current and original text

        deltaList = textDifference.getDelta(reader.text, originalText)

        correction = 0
        for ii in deltaList:
            # the correction is the difference in offset up to the position.
            position = ii[0]
            correction = ii[1]
            if bookmark.startOffset < position:
                # the positions of deltaList are in order. if our bookmark is below a position we have the correct correction
                break

        # add the correction to the list, correction = 0 if no correction was found
        reader._startOffset = bookmark.startOffset + correction
        reader._endOffset = bookmark.endOffset + correction
        return reader

    def onCallBack(self, index):
        text, info, startIndex, endIndex = textProcessor.sayAllWordDataBase[index]
        del textProcessor.sayAllWordDataBase[index]
        if not info:
            # we could not get the reader anymore
            if self.rectHighlighter:
                self.rectHighlighter.updateCustomRect(None)
            return
        
        info.collapse()
        info.move(textInfos.UNIT_CHARACTER, startIndex)
        info.move(textInfos.UNIT_CHARACTER, len(text), endPoint="end")

        if len(text) and text[-1] in [" ", "\n", "\r", "\t"]:
            text = text[:-1]
            info.move(textInfos.UNIT_CHARACTER, -1, endPoint="end")
        rect = textProcessor.getBoundingRects(info)
        # update the caret if applicable.
        if self.USA.getSettings().followWordJumpCaret:
            info.collapse()
            info.updateCaret()

        # highlight the word if applicable.
        if self.rectHighlighter:
            self.rectHighlighter.updateCustomRect(rect)

        # track to the rectangle
        self.USA.trackToRectangle(rect, context=Context.CARET)
